#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License
from typing import Iterable, List
import pytest
from itertools import permutations
import ezdxf.tools.text_layout as tl


@pytest.mark.parametrize(
    "margins,expected",
    [
        [None, (0, 0, 0, 0)],
        [(1,), (1, 1, 1, 1)],
        [(1, 2), (1, 2, 1, 2)],
        [(1, 2, 3), (1, 2, 3, 2)],
        [(1, 2, 3, 4), (1, 2, 3, 4)],
    ],
)
def test_resolve_margins(margins, expected):
    assert tl.resolve_margins(margins) == expected


@pytest.mark.parametrize(
    "align,expected",
    [
        [tl.LayoutAlignment.TOP_LEFT, (0, 0)],
        [tl.LayoutAlignment.TOP_CENTER, (-2, 0)],
        [tl.LayoutAlignment.TOP_RIGHT, (-4, 0)],
        [tl.LayoutAlignment.MIDDLE_LEFT, (0, 3)],
        [tl.LayoutAlignment.MIDDLE_CENTER, (-2, 3)],
        [tl.LayoutAlignment.MIDDLE_RIGHT, (-4, 3)],
        [tl.LayoutAlignment.BOTTOM_LEFT, (0, 6)],
        [tl.LayoutAlignment.BOTTOM_CENTER, (-2, 6)],
        [tl.LayoutAlignment.BOTTOM_RIGHT, (-4, 6)],
    ],
)
def test_insert_location(align, expected):
    assert tl.insert_location(align, width=4, height=6) == expected


class Rect(tl.ContentRenderer):
    def __init__(self, name: str, result: List = None):
        if result is None:
            result = []
        self.result = result  # store test results
        self.name = name

    def render(
        self, left: float, bottom: float, right: float, top: float, m=None
    ) -> None:
        self.result.append(
            f"{self.name}({left:.1f}, {bottom:.1f}, {right:.1f}, {top:.1f})"
        )

    def line(self, x1: float, y1: float, x2: float, y2: float, m=None) -> None:
        self.result.append(f"LINE({x1:.1f}, {y1:.1f})TO({x2:.1f}, {y2:.1f})")


class TestTopLevelLayout:
    @pytest.fixture
    def layout1(self):
        return tl.Layout(
            width=10, height=None, margins=(1, 1), renderer=Rect("Layout1")
        )

    def test_create_empty_layout_top_left(self, layout1):
        # layout1 has no height, only margins

        # 1. do layout placing
        layout1.place(align=tl.LayoutAlignment.TOP_LEFT)

        # 2. render content
        layout1.render()
        result = layout1.renderer.result
        assert len(result) == 1
        assert result[0] == "Layout1(0.0, -2.0, 12.0, 0.0)"

    def test_create_empty_layout_middle_center(self, layout1):
        # layout1 has no height, only margins

        # 1. do layout placing
        layout1.place(align=tl.LayoutAlignment.MIDDLE_CENTER)

        # 2. render content
        layout1.render()
        result = layout1.renderer.result
        assert len(result) == 1
        assert result[0] == "Layout1(-6.0, -1.0, 6.0, 1.0)"

    def test_add_one_column_by_reference_width(self, layout1):
        height = 17
        width = layout1.content_width  # reference column width
        result = layout1.renderer.result  # use same result container
        layout1.append_column(height=height, renderer=Rect("Col1", result))

        assert layout1.total_width == width + 2
        assert layout1.total_height == height + 2

        layout1.place(align=tl.LayoutAlignment.BOTTOM_LEFT)
        layout1.render()
        assert len(result) == 2
        assert result[0] == "Layout1(0.0, 0.0, 12.0, 19.0)"
        assert result[1] == "Col1(1.0, 1.0, 11.0, 18.0)"

    def test_add_two_equal_columns(self, layout1):
        margins = (1,)
        layout1.append_column(
            width=5, height=10, gutter=2, margins=margins, renderer=Rect("Col1")
        )
        layout1.append_column(
            width=7, height=20, margins=margins, renderer=Rect("Col2")
        )
        # width1 + margins + gutter + width2 + margins
        assert layout1.content_width == (5 + 2 + 2 + 7 + 2)

        # max(height) + margins
        assert layout1.content_height == (20 + 2)

    def test_bounding_box_for_not_placed_layout(self, layout1):
        # applies default alignment top/left, margins = (1, 1)
        layout1.append_column(10, 10)
        bbox = layout1.bbox()
        assert bbox.extmin == (0, -12)  # left/bottom
        assert bbox.extmax == (12, 0)  # right/top

    def test_bounding_box_for_placed_layout(self, layout1):
        # margins = (1, 1)
        layout1.append_column(10, 10)
        layout1.place(0, 0, tl.LayoutAlignment.MIDDLE_CENTER)
        bbox = layout1.bbox()
        assert bbox.extmin == (-6, -6)  # left/bottom
        assert bbox.extmax == (6, 6)  # right/top

    def test_next_existing_column(self, layout1):
        layout1.append_column(height=10)
        layout1.append_column(height=10)
        assert len(layout1) == 2
        assert layout1.current_column_index == 0
        layout1.next_column()
        assert layout1.current_column_index == 1

    def test_next_column_creates_a_new_column(self, layout1):
        layout1.append_column(height=10)
        assert len(layout1) == 1
        assert layout1.current_column_index == 0
        layout1.next_column()
        assert layout1.current_column_index == 1
        assert len(layout1) == 2, "a new column should be created"


class TestColumn:
    @pytest.fixture
    def c1(self):
        return tl.Column(
            # margins = top, right, bottom, left - same order as for CSS
            width=5,
            height=7,
            margins=(1, 2, 3, 4),
            renderer=Rect("C1"),
        )

    def test_size_calculation(self, c1):
        c1.place(0, 0)
        assert c1.content_width == 5
        assert c1.content_height == 7
        assert c1.total_width == 2 + 5 + 4
        assert c1.total_height == 1 + 7 + 3

    def test_render(self, c1):
        c1.place(0, 0)
        c1.render()
        result = c1.renderer.result
        assert result[0] == "C1(0.0, -11.0, 11.0, 0.0)"


def test_paragraph_available_line_content_space():
    par = tl.Paragraph(width=12, indent=(0.7, 0.5, 0.9))
    assert par.line_width(first=True) == 12 - 0.7 - 0.9
    assert par.line_width(first=False) == 12 - 0.5 - 0.9


class TestParagraphWithUnrestrictedHeight:
    # default values:
    # column width = 10
    # content width = 3
    # space width = 0.5

    @pytest.fixture
    def par(self):
        # Paragraph alignment is not important for content distribution,
        # because the required space is independent from alignment (left,
        # right, center or justified).
        # This may change by implementing regular tabulator support.
        return tl.Paragraph(width=10, renderer=Rect("PAR"))

    def test_empty_paragraph_dimensions(self, par):
        assert par.content_height == 0
        assert par.content_width == 10

    def test_render_empty_paragraph(self, par):
        par.place(0, 0)
        par.render()
        result = par.renderer.result

        assert len(result) == 1
        assert result[0] == "PAR(0.0, 0.0, 10.0, 0.0)"

    def test_distribute_invalid_content(self, par):
        par.append_content(str2cells("ttt"))
        with pytest.raises(tl.LayoutError):
            par.distribute_content(height=None)

    def test_distribute_common_case_without_nbsp(self, par):
        # column width = 10
        # content width = 3
        # space width = 0.5
        par.append_content(str2cells("t t t t t t t t t"))
        par.distribute_content(height=None)
        assert lines2str(par) == [
            "t t t",  # width = 3x3 + 2x0.5 = 10
            "t t t",  # remove line breaking spaces!
            "t t t",
        ]

    def test_distribute_with_nbsp(self, par):
        # column width = 10
        # content width = 3
        # space width = 0.5
        par.append_content(str2cells("t t t~t t t"))
        par.distribute_content(height=None)
        assert lines2str(par) == [
            "t t",  # t~t does not fit and goes to next line
            "t~t t",  # width = 3x3 + 2x0.5 = 10
            "t",
        ]

    def test_distribute_too_long_lines(self, par):
        # column width = 10
        par.append_content(str2cells("t t t", content=12))
        par.distribute_content(height=None)
        assert lines2str(par) == ["t", "t", "t"]

    def test_distribute_too_long_lines_including_nbsp(self, par):
        # column width = 10
        par.append_content(str2cells("t~t~t t~t t", content=5))
        par.distribute_content(height=None)
        assert lines2str(par) == [
            "t~t~t",  # width = 3x5 + 2x0.5 = 17
            "t~t",  # width = 2x5 + 0.5 = 10.5
            "t",
        ]


class TestParagraphWithRestrictedHeight:
    # default values:
    # column width = 10
    # content width = 3
    # space width = 0.5
    # cap height = 1,
    # line spacing 3-on-5 by 100% = 1.667
    THREE_LINE_SPACE = tl.leading(1, 1) * 2 + 1

    @pytest.fixture
    def par(self):
        # Paragraph alignment is not important for content distribution.
        return tl.Paragraph(width=10, renderer=Rect("PAR"))

    def test_distribute_with_exact_height_match(self, par):
        par.append_content(str2cells("t t t t t t t t t"))
        par.distribute_content(height=self.THREE_LINE_SPACE)
        assert lines2str(par) == [
            "t t t",  # width = 3x3 + 2x0.5 = 10
            "t t t",
            "t t t",
        ]

    def test_distribute_with_one_line_left_over(self, par):
        par.append_content(str2cells("t t t t t t t t t"))
        # Paragraph has only space for 2 lines by reducing the available space
        # by a small amount:
        height = self.THREE_LINE_SPACE - 0.01
        leftover = par.distribute_content(height=height)
        assert lines2str(par) == [
            "t t t",
            "t t t",
        ]
        leftover.distribute_content(height=1)
        assert lines2str(leftover) == ["t t t"]

    def test_distribute_with_all_lines_left_over(self, par):
        par.append_content(str2cells("t t t~t t t t t t"))
        # Paragraph has no space at all:
        leftover = par.distribute_content(height=0)
        assert lines2str(par) == []

        # None = unrestricted height
        leftover.distribute_content(height=None)
        assert lines2str(leftover) == [
            "t t",
            "t~t t",
            "t t t",
            "t",
        ]


def set_paragraph_content(flow):
    flow.append_content(str2cells("t t t t t t t t t"))
    flow.distribute_content()


class TestParagraphLeftAlignment:
    # default values:
    # content width = 3
    # space width = 0.5

    def test_without_indentation(self):
        par = tl.Paragraph(width=12, align=tl.ParagraphAlignment.LEFT)
        set_paragraph_content(par)
        par.place(0, 0)
        for line in par:
            assert line.total_width == 10
            assert line.final_location()[0] == 0

    def test_left_indentation(self):
        par = tl.Paragraph(
            width=12, indent=(0.7, 0.5, 0), align=tl.ParagraphAlignment.LEFT
        )
        set_paragraph_content(par)
        par.place(0, 0)
        lines = list(par)
        # first line:
        assert par.line_width(True) == 12 - 0.7  # available content space
        assert lines[0].final_location()[0] == 0.7
        assert lines[0].total_width == 10
        # remaining lines:
        for line in lines[1:]:
            assert par.line_width(False) == 12 - 0.5  # available content space
            assert line.total_width == 10
            assert line.final_location()[0] == 0.5

    def test_move_tab_to_next_line_if_following_content_does_not_fit(self):
        result = []
        par = tl.Paragraph(width=10, tab_stops=[tl.TabStop(4)])
        par.append_content(str2cells("t#t", content=6, result=result))
        # The tab (#) should move the following text to the tab stop
        # in the next line at position 4.
        par.distribute_content()
        par.place(0, 0)
        par.render()
        assert result[0] == "Text(0.0, -1.0, 6.0, 0.0)"
        assert result[1] == "Text(4.0, -2.7, 10.0, -1.7)", "x1 has to be 4.0"

    def test_tab_and_text_do_not_fit_into_line(self):
        result = []
        par = tl.Paragraph(width=6, tab_stops=[tl.TabStop(4)])
        par.append_content(str2cells("t#t", content=6, result=result))
        # The tabulator (#) and the following text move to the next line,
        # but also do not fit into a single line -> ignore tab
        par.distribute_content()
        par.place(0, 0)
        par.render()
        assert result[0] == "Text(0.0, -1.0, 6.0, 0.0)"
        assert result[1] == "Text(0.0, -2.7, 6.0, -1.7)", "x1 has to be 0.0"


class TestParagraphAlignment:
    # default values:
    # content width = 3
    # space width = 0.5

    def test_without_indentation(self):
        par = tl.Paragraph(width=12, align=tl.ParagraphAlignment.RIGHT)
        set_paragraph_content(par)
        par.place(0, 0)
        for line in par:
            assert line.total_width == 10
            assert line.final_location()[0] == 2

    def test_right_indentation(self):
        par = tl.Paragraph(
            width=12, indent=(0.5, 0.5, 0.5), align=tl.ParagraphAlignment.RIGHT
        )
        set_paragraph_content(par)
        par.place(0, 0)
        for line in par:
            assert line.total_width == 10
            assert line.final_location()[0] == 1.5  # 12 - 0.5 - 10


class TestParagraphCenterAlignment:
    # default values:
    # content width = 3
    # space width = 0.5

    def test_without_indentation(self):
        par = tl.Paragraph(width=12, align=tl.ParagraphAlignment.CENTER)
        set_paragraph_content(par)
        par.place(0, 0)
        for line in par:
            assert line.total_width == 10
            assert line.final_location()[0] == 1

    def test_left_indentation(self):
        par = tl.Paragraph(
            width=12, indent=(0.5, 0.5, 0), align=tl.ParagraphAlignment.CENTER
        )
        set_paragraph_content(par)
        par.place(0, 0)
        for line in par:
            assert line.total_width == 10
            assert line.final_location()[0] == 1.25  # 0.5 + (11.5 - 10) / 2

    def test_right_indentation(self):
        par = tl.Paragraph(
            width=12, indent=(0, 0, 0.5), align=tl.ParagraphAlignment.CENTER
        )
        set_paragraph_content(par)
        par.place(0, 0)
        for line in par:
            assert line.total_width == 10
            assert line.final_location()[0] == 0.75  # (11.5 - 10) / 2


class TestParagraphJustifiedAlignment:
    # default values:
    # content width = 3
    # space width = 0.5

    def test_without_indentation(self):
        par = tl.Paragraph(width=12, align=tl.ParagraphAlignment.JUSTIFIED)
        set_paragraph_content(par)
        par.place(0, 0)
        lines = list(par)
        for line in lines[:-1]:
            assert line.total_width == 12  # expand across paragraph width
            assert line.final_location()[0] == 0

        # last line is not expanded
        last_line = lines[-1]
        assert last_line.total_width == 10
        assert last_line.final_location()[0] == 0

    def test_with_indentation(self):
        par = tl.Paragraph(
            width=12,
            indent=(0.7, 0.5, 0.5),
            align=tl.ParagraphAlignment.JUSTIFIED,
        )
        set_paragraph_content(par)
        par.place(0, 0)
        lines = list(par)
        # first line:
        assert lines[0].total_width == 10.8  # 12 - (0.7 + 0.5)
        assert lines[0].final_location()[0] == 0.7

        # remaining lines:
        for line in lines[1:-1]:
            assert line.total_width == 11  # 12 - (0.5 + 0.5)
            assert line.final_location()[0] == 0.5

        # last line is not expanded:
        assert lines[-1].total_width == 10
        assert lines[-1].final_location()[0] == 0.5


class TestVerticalCellAlignment:
    @staticmethod
    def build_line(align):
        line = tl.LeftLine(width=7)
        big0 = tl.Text(width=3, height=3)
        small = tl.Text(width=1, height=1, valign=align, renderer=Rect("CELL"))
        big1 = tl.Text(width=3, height=3)
        line.append(big0)
        line.append(small)
        line.append(big1)
        line.place(0, 0)
        return line

    def test_line_properties(self):
        line = self.build_line(tl.CellAlignment.BOTTOM)
        assert len(list(line)) == 3
        assert line.total_width == 7
        assert line.total_height == 3

    def test_bottom_alignment(self):
        line = self.build_line(tl.CellAlignment.BOTTOM)
        big0, small, big1 = line
        # final location is always the top/left corner of the cell:
        assert big0.final_location() == (0, 0)
        assert small.final_location() == (3, -2)
        assert big1.final_location() == (4, 0)

        small.render()
        result = small.renderer.result
        # left, bottom, right, top
        assert result[0] == "CELL(3.0, -3.0, 4.0, -2.0)"

    def test_center_alignment(self):
        line = self.build_line(tl.CellAlignment.CENTER)
        big0, small, big1 = line
        # final location is always the top/left corner of the cell:
        assert big0.final_location() == (0, 0)
        assert small.final_location() == (3, -1)
        assert big1.final_location() == (4, 0)

        small.render()
        result = small.renderer.result
        # left, bottom, right, top
        assert result[0] == "CELL(3.0, -2.0, 4.0, -1.0)"

    def test_top_alignment(self):
        line = self.build_line(tl.CellAlignment.TOP)
        big0, small, big1 = line
        # final location is always the top/left corner of the cell:
        assert big0.final_location() == (0, 0)
        assert small.final_location() == (3, 0)
        assert big1.final_location() == (4, 0)

        small.render()
        result = small.renderer.result
        # left, bottom, right, top
        assert result[0] == "CELL(3.0, -1.0, 4.0, 0.0)"

    def test_mixed_alignment(self):
        big0 = tl.Text(width=3, height=3)
        bottom = tl.Text(width=1, height=1, valign=tl.CellAlignment.BOTTOM)
        center = tl.Text(width=1, height=1, valign=tl.CellAlignment.CENTER)
        top = tl.Text(width=1, height=1, valign=tl.CellAlignment.TOP)
        big1 = tl.Text(width=3, height=3)
        line = tl.LeftLine(width=9)
        for cell in [big0, top, center, bottom, big1]:
            line.append(cell)
        line.place(0, 0)
        assert bottom.final_location() == (5, -2)
        assert center.final_location() == (4, -1)
        assert top.final_location() == (3, 0)


class StrokeRender(Rect):
    def line(self, x1: float, y1: float, x2: float, y2: float, m=None) -> None:
        length = x2 - x1
        if y1 < -1:
            location = "UNDERLINE"
        elif y1 > 0:
            location = "OVERLINE"
        else:
            location = "STRIKE_THROUGH"
        self.result.append(f"{self.name}({location}, {length:.1f})")


class TestTextStrokeRendering:
    @staticmethod
    def render_text(stroke, result):
        text = tl.Text(
            width=3,
            height=1,
            stroke=stroke,
            renderer=StrokeRender("STROKE", result),
        )
        text.place(0, 0)
        tl.render_text_strokes([text])

    @pytest.mark.parametrize(
        "stroke,expected",
        [
            (tl.Stroke.UNDERLINE, "STROKE(UNDERLINE, 3.0)"),
            (tl.Stroke.OVERLINE, "STROKE(OVERLINE, 3.0)"),
            (tl.Stroke.STRIKE_THROUGH, "STROKE(STRIKE_THROUGH, 3.0)"),
        ],
    )
    def test_simple_stroke(self, stroke, expected):
        result = []
        self.render_text(stroke, result)
        assert result[0] == expected


class TestTextContinueStroke:
    @staticmethod
    def make_text(stroke, result):
        text = tl.Text(
            width=3,
            height=1,
            stroke=stroke,
            renderer=StrokeRender("STROKE", result),
        )
        text.place(0, 0)
        return text

    def test_continue_stroke_across_one_space(self):
        result = []
        word = self.make_text(tl.Stroke.UNDERLINE + tl.Stroke.CONTINUE, result)
        space = tl.Space(width=0.5)
        tl.render_text_strokes([word, space, word])
        assert len(result) == 2
        assert result[0] == "STROKE(UNDERLINE, 3.5)", "space should be included"
        assert result[1] == "STROKE(UNDERLINE, 3.0)", "no following space"

    def test_continue_stroke_across_multiple_spaces(self):
        result = []
        word = self.make_text(tl.Stroke.UNDERLINE + tl.Stroke.CONTINUE, result)
        space = tl.Space(width=0.5)
        nbsp = tl.NonBreakingSpace(width=0.5)
        tl.render_text_strokes([word, space, nbsp, space, word])
        assert len(result) == 2
        assert result[0] == "STROKE(UNDERLINE, 4.5)", "3 spaces should be included"
        assert result[1] == "STROKE(UNDERLINE, 3.0)", "no following spaces"

    def test_do_not_continue_stroke_automatically(self):
        result = []
        word = self.make_text(tl.Stroke.UNDERLINE, result)
        space = tl.Space(width=0.5)
        tl.render_text_strokes([word, space, word])
        assert len(result) == 2
        assert result[0] == "STROKE(UNDERLINE, 3.0)", "do not continue stroke"


class TestFractionCell:
    @staticmethod
    def fraction(stacking, x, y):
        result = []
        a = tl.Text(1, 1, renderer=Rect("A", result))
        b = tl.Text(1, 1, renderer=Rect("B", result))
        fr = tl.Fraction(a, b, stacking, renderer=Rect("Fraction", result))
        fr.place(x, y)
        fr.render()
        return result

    def test_a_over_b(self):
        # y = total height = (a.total_height + b.total_height) * HEIGHT_SCALE
        result = self.fraction(tl.Stacking.OVER, x=0, y=2 * tl.Fraction.HEIGHT_SCALE)
        assert len(result) == 2
        assert result[0] == "A(0.0, 1.4, 1.0, 2.4)"  # L, B, R, T
        assert result[1] == "B(0.0, 0.0, 1.0, 1.0)"  # L, B, R, T

    def test_a_over_line_b(self):
        # y = total height = (a.total_height + a.total_height) * HEIGHT_SCALE
        result = self.fraction(tl.Stacking.LINE, x=0, y=2 * tl.Fraction.HEIGHT_SCALE)
        assert len(result) == 3
        assert result[0] == "A(0.0, 1.4, 1.0, 2.4)"  # L, B, R, T
        assert result[1] == "B(0.0, 0.0, 1.0, 1.0)"  # L, B, R, T
        assert result[2] == "LINE(0.0, 1.2)TO(1.0, 1.2)"

    def test_a_slanted_b(self):
        # y = total height = (a.total_height + a.total_height)
        result = self.fraction(tl.Stacking.SLANTED, x=0, y=2)
        assert len(result) == 3
        assert result[0] == "A(0.0, 1.0, 1.0, 2.0)"  # L, B, R, T
        assert result[1] == "B(1.0, 0.0, 2.0, 1.0)"  # L, B, R, T
        assert result[2] == "LINE(0.0, 0.0)TO(2.0, 2.0)"


def str2cells(
    s: str, content: float = 3, space: float = 0.5, tab: float = 0, result=None
):
    # t ... text cell
    # f ... fraction cell
    # space is space
    # ~ ... non breaking space (nbsp)
    # # ... tabulator
    if result is None:
        result = []
    for c in s.lower():
        if c == "t":
            yield tl.Text(width=content, height=1, renderer=Rect("Text", result=result))
        elif c == "f":
            cell = tl.Text(content / 2, 1)
            yield tl.Fraction(
                top=cell,
                bottom=cell,
                stacking=tl.Stacking.SLANTED,
                renderer=Rect("Fraction", result=result),
            )
        elif c == " ":
            yield tl.Space(width=space)
        elif c == "~":
            yield tl.NonBreakingSpace(width=space)
        elif c == "#":
            yield tl.Tabulator(width=tab)  # Tabulators do not need a width
        else:
            raise ValueError(f'unknown cell type "{c}"')


CELL2STR = {
    tl.Text: "t",
    tl.Fraction: "f",
    tl.Space: " ",
    tl.NonBreakingSpace: "~",
    tl.Tabulator: "#",
}


def cells2str(cells: Iterable[tl.Cell]) -> str:
    return "".join(CELL2STR[type(cell)] for cell in cells)


def lines2str(lines):
    return [cells2str(line) for line in lines]


def test_cell_converter():
    assert cells2str(str2cells("tf ~#")) == "tf ~#"
    with pytest.raises(ValueError):
        list(str2cells("x"))
    with pytest.raises(KeyError):
        cells2str([0])


class TestNormalizeCells:
    @pytest.mark.parametrize("content", ["tt", "tf", "ft", "ff"])
    def test_no_glue_between_content_raises_value_error(self, content):
        cells = str2cells(content)
        with pytest.raises(tl.LayoutError):
            list(tl.normalize_cells(cells))

    @pytest.mark.parametrize("content", ["t~f", "f~f", "f~t"])
    def test_ignore_non_breaking_space_between_text_and_fraction(self, content):
        cells = str2cells(content)
        result = tl.normalize_cells(cells)
        assert len(result) == 3

    def test_ignore_pending_non_breaking_space(self):
        cells = str2cells("t~t~")
        result = tl.normalize_cells(cells)
        assert len(result) == 3

    @pytest.mark.parametrize("content", ["t~t", "t~~t", "t~~~t"])
    def test_preserve_multiple_nbsp(self, content):
        cells = tl.normalize_cells(str2cells(content))
        assert cells2str(cells) == content

    @pytest.mark.parametrize(
        "content",
        [
            "t~ t",
            "t ~t",
            "t~~ t",
            "t ~~t",
            "~t",
            "~~t",
            "t#~t",
            "t~#t",
            "t~#~t",
        ],
    )
    def test_replace_useless_nbsp_by_spaces(self, content):
        cells = tl.normalize_cells(str2cells(content))
        assert cells2str(cells) == content.replace("~", " ")

    @pytest.mark.parametrize("content", ["t t", "t  t", "t   t"])
    def test_preserve_multiple_spaces(self, content):
        cells = tl.normalize_cells(str2cells(content))
        assert cells2str(cells) == content

    def test_remove_pending_glue(self):
        for glue in permutations([" ", "~", " ", "#"]):
            content = "t" + "".join(glue)
            cells = list(tl.normalize_cells(str2cells(content)))
            assert cells2str(cells) == "t"

    @pytest.mark.parametrize("content", [" t", "  t", "   t"])
    def test_preserve_prepending_space(self, content):
        cells = list(tl.normalize_cells(str2cells(content)))
        assert cells2str(cells) == content


class TestSpace:
    def test_shrink_space(self):
        space = tl.Space(1, min_width=0.1)
        space.resize(0.5)
        assert space.total_width == 0.5
        space.resize(0)
        assert space.total_width == 0.1

    def test_default_min_width(self):
        space = tl.Space(1)
        space.resize(0.5)
        assert space.total_width == 1.0

    def test_expand_restricted_space(self):
        space = tl.Space(1, max_width=2)
        space.resize(1.5)
        assert space.total_width == 1.5
        space.resize(3)
        assert space.total_width == 2

    def test_expand_unrestricted_space(self):
        space = tl.Space(1)
        space.resize(1.5)
        assert space.total_width == 1.5
        space.resize(30)
        assert space.total_width == 30

    def test_total_height_is_zero(self):
        assert tl.Space(1).total_height == 0

    def test_non_breaking_space_to_space(self):
        space = tl.NonBreakingSpace(1).to_space()
        assert type(space) == tl.Space

    def test_can_shrink(self):
        assert tl.Space(1).can_shrink is False
        assert tl.Space(1, min_width=0.5).can_shrink is True

    def test_can_grow(self):
        assert tl.Space(1).can_grow is True
        assert tl.Space(1, max_width=1.0).can_grow is False


class TestRigidConnection:
    def test_rigid_connection(self):
        cells = tl.normalize_cells(str2cells("t~t t t"))
        result = tl.group_non_breakable_cells(cells)
        assert isinstance(result[0], tl.RigidConnection)
        assert isinstance(result[1], tl.Space)
        assert isinstance(result[2], tl.Text)
        assert isinstance(result[3], tl.Space)
        assert isinstance(result[4], tl.Text)

    @pytest.mark.parametrize("content", ["t~t", "t~t~t"])
    def test_create_one_connection(self, content):
        cells = tl.normalize_cells(str2cells(content))
        result = tl.group_non_breakable_cells(cells)
        assert len(result) == 1

    @pytest.mark.parametrize("content", ["t~t t~t", "t~t~t t~t~t"])
    def test_create_two_connections(self, content):
        cells = tl.normalize_cells(str2cells(content))
        result = tl.group_non_breakable_cells(cells)
        assert len(result) == 3
        assert isinstance(result[1], tl.Space)

    def test_ignore_pending_non_breaking_space(self):
        cells = tl.normalize_cells(str2cells("t~t~"))
        result = tl.group_non_breakable_cells(cells)
        assert len(result) == 1


def render_line_with_tabs(
    line: tl.AbstractLine,
    cells: Iterable[tl.Cell],
):
    result = []
    tab = None
    tab_renderer = Rect("TAB-TEXT", result=result)
    text_renderer = Rect("TEXT", result=result)
    for cell in cells:
        cell.renderer = text_renderer
        if tab is not None:
            cell.renderer = tab_renderer
            line.append_with_tab(cell, tab)
            tab = None
            continue
        if isinstance(cell, tl.Tabulator):
            tab = cell
        else:
            tab = None
            line.append(cell)
    line.place(0, 0)
    line.render()
    return result


class TestLeftLine:
    def test_setup(self):
        line = tl.LeftLine(10)
        assert line.line_width == 10
        assert line.total_width == 0

    def test_line_height_is_defined_by_max_content_height(self):
        line = tl.LeftLine(10)
        line.append(tl.Text(1, 1))
        assert line.total_height == 1
        line.append(tl.Text(1, 2))
        assert line.total_height == 2
        line.append(tl.Text(1, 3))
        assert line.total_height == 3

    def test_line_total_width_is_defined_by_content(self):
        line = tl.LeftLine(10)
        line.append(tl.Text(1, 1))
        assert line.total_width == 1
        line.append(tl.Text(1, 1))
        assert line.total_width == 2
        line.append(tl.Text(1, 1))
        assert line.total_width == 3

    def test_fill_until_line_is_full(self):
        line = tl.LeftLine(10)
        assert line.append(tl.Text(5, 1)) == tl.AppendType.SUCCESS
        assert line.append(tl.Text(5, 1)) == tl.AppendType.SUCCESS
        assert line.append(tl.Text(5, 1)) == tl.AppendType.FAIL
        assert line.total_width <= line.line_width

    def test_left_tab(self):
        line = tl.LeftLine(
            width=20,
            tab_stops=[
                tl.TabStop(6, tl.TabStopType.LEFT),
                tl.TabStop(12, tl.TabStopType.LEFT),
            ],
        )
        cells = str2cells("t#t#t t", content=2, space=0.5)
        result = render_line_with_tabs(line, cells)
        assert result[0] == "TEXT(0.0, -1.0, 2.0, 0.0)"
        assert result[1] == "TAB-TEXT(6.0, -1.0, 8.0, 0.0)"
        assert result[2] == "TAB-TEXT(12.0, -1.0, 14.0, 0.0)"
        assert result[3] == "TEXT(14.5, -1.0, 16.5, 0.0)"

    def test_left_tab_without_tab_stops(self):
        line = tl.LeftLine(20)
        cells = str2cells("t#t", content=2, space=0.5, tab=0.5)
        result = render_line_with_tabs(line, cells)
        assert result[0] == "TEXT(0.0, -1.0, 2.0, 0.0)"
        # replace tab by a space width = 0.5
        assert result[1] == "TAB-TEXT(2.5, -1.0, 4.5, 0.0)"

    def test_center_tab(self):
        line = tl.LeftLine(
            width=20,
            tab_stops=[
                tl.TabStop(6, tl.TabStopType.CENTER),
                tl.TabStop(12, tl.TabStopType.CENTER),
            ],
        )
        cells = str2cells("t#t#t t", content=2, space=0.5)
        result = render_line_with_tabs(line, cells)
        assert result[0] == "TEXT(0.0, -1.0, 2.0, 0.0)"
        assert result[1] == "TAB-TEXT(5.0, -1.0, 7.0, 0.0)"
        assert result[2] == "TAB-TEXT(11.0, -1.0, 13.0, 0.0)"
        assert result[3] == "TEXT(13.5, -1.0, 15.5, 0.0)"

    def test_right_tab(self):
        line = tl.LeftLine(
            width=20,
            tab_stops=[
                tl.TabStop(6, tl.TabStopType.RIGHT),
                tl.TabStop(12, tl.TabStopType.RIGHT),
            ],
        )
        cells = str2cells("t#t#t t", content=2, space=0.5)
        result = render_line_with_tabs(line, cells)
        assert result[0] == "TEXT(0.0, -1.0, 2.0, 0.0)"
        assert result[1] == "TAB-TEXT(4.0, -1.0, 6.0, 0.0)"
        assert result[2] == "TAB-TEXT(10.0, -1.0, 12.0, 0.0)"
        assert result[3] == "TEXT(12.5, -1.0, 14.5, 0.0)"


def tab_stops(*values):
    return [tl.TabStop(v) for v in values]


def test_shift_tab_stop_left_in_range():
    result = tl.shift_tab_stops(tab_stops(4, 8), -2, 10)
    assert result[0].pos == 2
    assert result[1].pos == 6


def test_shift_tab_stops_beyond_left_border():
    result = tl.shift_tab_stops(tab_stops(4, 8), -4, 10)
    assert len(result) == 1, "tab stop at pos=0 should be removed"
    assert result[0].pos == 4


def test_shift_tab_stops_beyond_right_border():
    result = tl.shift_tab_stops(tab_stops(4, 8), 4, 8)
    assert len(result) == 1
    assert result[0].pos == 8, "tab stop at the right border should be preserved"


def test_empty_paragraph():
    cap_height = 2
    p = tl.EmptyParagraph(cap_height, 1)
    assert p.total_width == 0
    assert p.total_height == cap_height
    assert p.distance_to_next_paragraph > cap_height / 2


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2022, Manfred Moitzi
# License: MIT License
""" ezdxf typing hints

Only usable in type checking mode:

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.eztypes import GenericLayoutType

Tips for Type Imports
---------------------

Import Drawing class:

    from ezdxf.document import Drawing

Import DXF entities from ezdxf.entities:

    from ezdxf.entities import Line, Point, ...

Import layouts from ezdxf.layouts:

   from ezdxf.layouts import BaseLayout, Layout, Modelspace, Paperspace, BlockLayout

Import math tools from ezdxf.math:

    from ezdxf.math import Vec2, Vec3, Matrix44, ...

Import path tools from ezdxf.path:

    from ezdxf.path import Path, make_path, ...

"""
from __future__ import annotations
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Sequence,
    TYPE_CHECKING,
    Tuple,
    Union,
)
from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from ezdxf.entities import DXFEntity
    from ezdxf.layouts.base import VirtualLayout
    from ezdxf.layouts.blocklayout import BlockLayout
    from ezdxf.layouts.layout import Layout
    from ezdxf.lldxf.extendedtags import ExtendedTags
    from ezdxf.lldxf.tags import Tags
    from ezdxf.math import UVec

    IterableTags: TypeAlias = Iterable[Tuple[int, Any]]
    SectionDict: TypeAlias = Dict[str, List[Union[Tags, ExtendedTags]]]
    KeyFunc: TypeAlias = Callable[[DXFEntity], Hashable]
    FaceType: TypeAlias = Sequence[UVec]
    GenericLayoutType: TypeAlias = Union[Layout, BlockLayout, VirtualLayout]



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import Vec2, BoundingBox2d
from ezdxf.math.clipping import (
    greiner_hormann_union,
    greiner_hormann_difference,
    greiner_hormann_intersection,
    line_intersection,
)
from ezdxf.render.forms import circle


class TestLineIntersection:
    """The Greiner/Hormann algorithm needs a special intersection function.

    Start and end points of a line are not intersection points!

    Which also leads to issues for joining adjacent but not intersecting
    polygons.

    """

    def test_start_point_is_not_an_intersection_point(self):
        s1, s2 = Vec2(1, 1), Vec2(3, 1)
        c1, c2 = Vec2(2, 1), Vec2(2, 2)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    def test_end_point_is_not_an_intersection_point(self):
        s1, s2 = Vec2(1, 1), Vec2(3, 1)
        c1, c2 = Vec2(2, 2), Vec2(2, 1)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    def test_corner_point_is_not_an_intersection_point(self):
        s1, s2 = Vec2(1, 1), Vec2(3, 1)
        c1, c2 = Vec2(1, 1), Vec2(1, 2)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    def test_vertical_line_does_intersect_skewed_line(self):
        s1, s2 = Vec2(10, 1), Vec2(10, -7)
        c1, c2 = Vec2(-10, 3), Vec2(17, -7)
        point, *_ = line_intersection(s1, s2, c1, c2)
        assert point.x == 10
        assert point.isclose(Vec2(10.0, -4.4074), abs_tol=1e-4)

    def test_horizontal_line_does_intersect_skewed_line(self):
        s1, s2 = Vec2(-10, 10), Vec2(10, 10)
        c1, c2 = Vec2(-10, 20), Vec2(10, 0)
        point, *_ = line_intersection(s1, s2, c1, c2)
        assert point.y == 10
        assert point.isclose(Vec2(0.0, 10.0), abs_tol=1e-4)

    def test_orthogonal_lines_do_intersect(self):
        s1, s2 = Vec2(-10, 10), Vec2(10, 10)
        c1, c2 = Vec2(5, 0), Vec2(5, 20)
        point, us, uc = line_intersection(s1, s2, c1, c2)
        assert point.y == 10
        assert point.x == 5
        assert point.isclose(Vec2(5.0, 10.0), abs_tol=1e-4)
        assert us == pytest.approx(0.75)
        assert uc == pytest.approx(0.5)

    def test_parallel_vertical_lines_do_not_intersect(self):
        s1, s2 = Vec2(10, 1), Vec2(10, -7)
        c1, c2 = Vec2(12, -10), Vec2(12, 7)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    def test_parallel_horizontal_lines_do_not_intersect(self):
        s1, s2 = Vec2(11, 0), Vec2(-11, 0)
        c1, c2 = Vec2(0, 0), Vec2(1, 0)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    def test_collinear_lines_do_not_intersect(self):
        s1, s2 = Vec2(0, 0), Vec2(2, 2)
        c1, c2 = Vec2(3, 3), Vec2(4, 4)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    @pytest.mark.parametrize(
        "p2", [(4, 0), (0, 4), (4, 4)], ids=["horiz", "vert", "diag"]
    )
    def test_coincident_lines_do_not_intersect(self, p2):
        s1, s2 = Vec2(0, 0), Vec2(p2)
        assert line_intersection(s1, s2, s1, s2)[0] is None

    def test_virtual_intersection_is_not_an_intersection(self):
        s1, s2 = Vec2(0, 0), Vec2(4, 4)
        c1, c2 = Vec2(3, 2), Vec2(5, 0)
        assert line_intersection(s1, s2, c1, c2)[0] is None

    def test_issue_128(self):
        s1, s2 = Vec2(175.0, 5.0), Vec2(175.0, 50.0)
        c1, c2 = Vec2(-10.1231, 30.1235), Vec2(300.2344, 30.1235)
        point, *_ = line_intersection(s1, s2, c1, c2)
        assert point.isclose(Vec2(175.0, 30.1235))

    def test_issue_664(self):
        s1 = Vec2(16399619.87946683, -199438.8133075837)
        s2 = Vec2(16399700.34999472, -199438.8133075837)
        c1 = Vec2(16399659.76235549, -199423.8133075837)
        c2 = Vec2(16399659.76235549, -199463.8133075837)
        point, *_ = line_intersection(s1, s2, c1, c2)
        assert point.isclose(Vec2(16399659.76235549, -199438.8133075837))


@pytest.fixture
def rect():
    return Vec2.list([(-1, -1), (1, -1), (1, 1), (-1, 1)])


@pytest.fixture
def overlapping():  # overlapping
    return Vec2.list([(0, 0), (2, 0), (2, 2), (0, 2)])


@pytest.fixture
def inside():  # complete inside
    return Vec2.list([(0, 0), (0.5, 0), (0.5, 0.5), (0, 0.5)])


@pytest.fixture
def outside():  # complete outside
    return Vec2.list([(2, 2), (3, 2), (3, 3), (2, 3)])


UNION_OVERLAPPING = Vec2.list(
    [
        (-1, -1),
        (1, -1),
        (1, 0),
        (2, 0),
        (2, 2),
        (0, 2),
        (0, 1),
        (-1, 1),
    ]
)

UNION_OUTSIDE = Vec2.list(
    [
        (-1, -1),
        (1, -1),
        (1, 1),
        (2, 0),
        (2, 2),
        (0, 2),
        (0, 1),
        (-1, 1),
    ]
)


class TestBooleanUnion:
    def test_overlapping_polygons_are_united(self, rect, overlapping):
        """The existence of real intersection points is the basic requirement
        to get the union operator to work.
        """
        polygons = greiner_hormann_union(rect, overlapping)
        assert len(polygons) == 1
        result = polygons[0]
        assert len(result) == 9
        assert result[0] == result[-1], "expected closed polygon"
        assert set(UNION_OVERLAPPING) == set(result)

    def test_vertex_order_is_not_important(self, rect, overlapping):
        """The vertex order (clockwise or counter clockwise) is not important."""
        rect.reverse()
        result = greiner_hormann_union(rect, overlapping)[0]
        assert set(UNION_OVERLAPPING) == set(result)

        overlapping.reverse()
        result = greiner_hormann_union(rect, overlapping)[0]
        assert set(UNION_OVERLAPPING) == set(result)

    def test_a_polygon_inside_another_polygon_is_ignored(self, rect, inside):
        """This polygons do not have any intersection points and therefore no
        union is calculated.
        """
        assert len(greiner_hormann_union(rect, inside)) == 0

    def test_a_failed_union_returns_an_empty_list(self, rect, outside):
        assert len(greiner_hormann_union(rect, outside)) == 0

    def test_disconnected_polygons_cannot_be_united(self, rect):
        assert len(greiner_hormann_union(rect, circle(16, 3))) == 0

    def test_overlapping_but_collinear_edges_cannot_be_united(self, rect):
        """As shown in the intersection tests, collinear lines do not intersect,
        and this algorithm relies on intersections to produce a union.
        """
        offset = Vec2(1.5, 0)
        rect2 = [offset + v for v in rect]
        assert len(greiner_hormann_union(rect, rect2)) == 0

    def test_polygons_with_a_shared_edge_cannot_be_united(self, rect):
        """As shown in the intersection tests, collinear lines do not intersect,
        and this algorithm relies on intersections to produce a union.
        """
        offset = Vec2(2, 0)
        rect2 = [offset + v for v in rect]
        assert len(greiner_hormann_union(rect, rect2)) == 0


DIFF_OVERLAPPING = [
    Vec2(-1, -1),
    Vec2(1, -1),
    Vec2(1, 0),
    Vec2(0, 0),
    Vec2(0, 1),
    Vec2(-1, 1),
]


class TestBooleanDifference:
    def test_difference_of_overlapping_polygons(self, rect, overlapping):
        """The existence of real intersection points is the basic requirement
        to get the difference operator to work.
        """
        polygons = greiner_hormann_difference(rect, overlapping)
        assert len(polygons) == 1
        result = polygons[0]
        assert len(result) == 7
        assert result[0] == result[-1], "expected closed polygon"
        assert set(DIFF_OVERLAPPING) == set(result)

    def test_polygon_inside_polygon(self, rect, inside):
        """This polygons do not have any intersection points and therefore no
        difference is calculated - which is NOT the expected result
        (expected a rect with a hole)
        """
        assert len(greiner_hormann_difference(rect, inside)) == 0

    def test_polygon_inside_polygon_reverse_difference(self, rect, inside):
        """This polygons do not have any intersection points and therefore no
        difference is calculated - which IS the expected result
        (subtracting the bigger rect from the smaller rect returns nothing)
        """
        assert len(greiner_hormann_difference(inside, rect)) == 0

    def test_polygon_outside_polygon(self, rect, outside):
        """This polygons do not have any intersection points and therefore no
        difference is calculated - which is NOT the expected result
        (expected rect)
        """
        assert len(greiner_hormann_difference(rect, outside)) == 0


INTERSECT_OVERLAPPING = [
    Vec2(0, 0),
    Vec2(1, 0),
    Vec2(1, 1),
    Vec2(0, 1),
]


class TestBooleanIntersection:
    def test_intersection_of_overlapping_polygons(self, rect, overlapping):
        """The existence of real intersection points is the basic requirement
        to get the intersection operator to work.
        """
        polygons = greiner_hormann_intersection(rect, overlapping)
        assert len(polygons) == 1
        result = polygons[0]
        assert len(result) == 5
        assert result[0] == result[-1], "expected closed polygon"
        assert set(INTERSECT_OVERLAPPING) == set(result)

    def test_polygon_inside_polygon(self, rect, inside):
        """This polygons do not have any intersection points and therefore no
        intersection is calculated - which is NOT the expected result
        (expected the inside rect)
        """
        assert len(greiner_hormann_intersection(rect, inside)) == 0

    def test_polygon_outside_polygon(self, rect, outside):
        """This polygons do not have any intersection points and therefore no
        intersection is calculated - which could interpreted as the expected
        result.
        """
        assert len(greiner_hormann_intersection(rect, outside)) == 0


DATA_1094 = [
    Vec2.list(
        [
            [-18.900001889999952, 1450.7000002308125],
            [-18.900001889999952, 1470.7351470734998],
            [1278.90012789, 1470.7351470734998],
            [1278.90012789, 1450.70001562087],
            [1503.6001841952093, 1450.7000156208703],
            [1503.6001841952093, 1388.3850000081884],
            [1503.6001841952093, 1388.3850000081884],
            [1647.3246376764234, 1388.3850000081884],
            [1647.3246376764234, 1548.1475422037274],
            [1831.0471719722357, 1548.1475422037274],
            [1831.0471719722357, 1388.3850000081884],
            [1831.0471719722357, 1388.3850000081884],
            [1980.6216566984945, 1388.3850000081884],
            [1980.6216566984945, 1224.6149999918116],
            [1831.0471719722357, 1224.6149999918116],
            [1831.0471719722357, 1160.3525263520064],
            [1647.3246376764234, 1160.3525263520064],
            [1647.3246376764234, 1160.3525263520064],
            [1647.3246376764234, 1224.6149999918116],
            [1647.3246376764234, 1224.6149999918116],
            [1503.6001841952093, 1224.6149999918116],
            [1503.6001841952093, 1162.2999997691875],
            [1278.90012789, 1162.2999997691877],
            [1278.90012789, 1162.2999997691877],
            [1278.90012789, 286.7000156208702],
            [1503.6001841952093, 286.70001562087026],
            [1503.6001841952093, 224.3850000081885],
            [1503.6001841952093, 224.3850000081885],
            [1647.3246376764234, 224.3850000081885],
            [1647.3246376764234, 384.1475422037275],
            [1831.0471719722357, 384.1475422037275],
            [1831.0471719722357, 224.3850000081885],
            [1831.0471719722357, 224.3850000081885],
            [1980.6216566984945, 224.3850000081885],
            [1980.6216566984945, 60.61499999181149],
            [1831.0471719722357, 60.614999991811544],
            [1831.0471719722357, -3.6474736479937064],
            [1647.3246376764234, -3.6474736479937064],
            [1647.3246376764234, -3.6474736479937064],
            [1647.3246376764234, 60.61499999181149],
            [1647.3246376764234, 60.61499999181149],
            [1503.6001841952093, 60.61499999181149],
            [1503.6001841952093, -1.7000002308124635],
            [1278.90012789, -1.7000002308124635],
            [1278.90012789, -1.7000002308124635],
            [1278.90012789, -21.735002173499993],
            [-18.900001890000045, -21.735002173499993],
            [-18.900001890000045, -21.735002173499993],
            [-18.900001890000045, -1.7000156208705448],
            [-243.6001841952094, -1.7000156208702606],
            [-243.6001841952094, -1.7000156208702606],
            [-243.6001841952094, 60.61499999181149],
            [-243.6001841952094, 60.61499999181149],
            [-237.7502641782885, 60.61499999181149],
            [-237.7502641782885, 224.3850000081885],
            [-243.6001841952094, 224.3850000081885],
            [-243.6001841952094, 286.70000023081246],
            [-18.900001890000027, 286.7000002308125],
            [-18.900001890000027, 286.7000002308125],
            [-18.90000188999997, 1162.2999843791297],
            [-243.6001841952094, 1162.2999843791297],
            [-243.6001841952094, 1162.2999843791297],
            [-243.6001841952094, 1450.7000002308125],
            [-18.900001889999952, 1450.7000002308125],
        ]
    ),
    Vec2.list(
        [
            [-392.2502641782885, 1224.6149999918116],
            [-237.7502641782885, 1224.6149999918116],
            [-237.7502641782885, 1388.3850000081884],
            [-392.2502641782885, 1388.3850000081884],
            [-392.2502641782885, 1224.6149999918116],
        ]
    ),
]


def test_issue_1094_is_inside_polygon_function_was_incorrect():
    data = DATA_1094
    box_expected = BoundingBox2d(data[0] + data[1])
    result = greiner_hormann_union(data[0], data[1])

    assert len(result) == 1
    box_result = BoundingBox2d(result[0])
    assert box_result.extmin.isclose(box_expected.extmin)
    assert box_result.extmax.isclose(box_expected.extmax)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2014-2024, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

from ezdxf.entities.acis import tags2textlines, textlines2tags


@pytest.fixture(scope="module")
def layout():
    doc = ezdxf.new("R2007")
    return doc.modelspace()


def test_body_default_settings(layout):
    body = layout.add_body()
    assert "0" == body.dxf.layer


def test_body_getting_acis_data(layout):
    body = layout.add_body()
    body.sat = TEST_DATA.splitlines()
    assert TEST_DATA == body.tostring()


def test_region_default_settings(layout):
    region = layout.add_region()
    assert region.dxf.layer == "0"


def test_region_getting_acis_data(layout):
    region = layout.add_region()
    region.sat = TEST_DATA.splitlines()
    assert TEST_DATA == region.tostring()


def test_3dsolid_default_settings(layout):
    _3dsolid = layout.add_3dsolid()
    assert _3dsolid.dxf.layer == "0"
    assert _3dsolid.dxf.history_handle == "0"


def test_3dsolid_getting_acis_data(layout):
    _3dsolid = layout.add_3dsolid()
    _3dsolid.sat = TEST_DATA.splitlines()
    assert TEST_DATA == _3dsolid.tostring()


TEST_DATA = """21200 115 2 26
16 Autodesk AutoCAD 19 ASM 217.0.0.4503 NT 0
1 9.9999999999999995e-007 1e-010
asmheader $-1 -1 @12 217.0.0.4503 #
body $2 -1 $-1 $3 $-1 $-1 #
ref_vt-eye-attrib $-1 -1 $-1 $-1 $1 $4 $5 #
lump $6 -1 $-1 $-1 $7 $1 #
eye_refinement $-1 -1 @5 grid  1 @3 tri 1 @4 surf 0 @3 adj 0 @4 grad 0 @9 postcheck 0 @4 stol 0.020115179941058159 @4 ntol 30 @4 dsil 0 @8 flatness 0 @7 pixarea 0 @4 hmax 0 @6 gridar 0 @5 mgrid 3000 @5 ugrid 0 @5 vgrid 0 @10 end_fields #
vertex_template $-1 -1 3 0 1 8 #"""


def test_tag2lines():
    expected = "AB" * 50 + "CD" * 50 + "EF" * 50
    tags = [
        (1, "AB" * 50),
        (3, "CD" * 50),
        (3, "EF" * 50),
    ]
    assert list(tags2textlines(tags))[0] == expected


def test_lines2tags():
    line = "AB" * 100 + "CD" * 100 + "EF" * 100
    result = list(textlines2tags([line]))
    code, value = result[0]
    assert code == 1
    assert value == line[:255]
    code, value = result[1]
    assert code == 3
    assert value == line[255:510]
    code, value = result[2]
    assert code == 3
    assert value == line[510:]



# Copyright (c) 2022, Manfred Moitzi
# License: MIT License
# This add-on was created to solve this problem: https://github.com/mozman/ezdxf/discussions/789
from __future__ import annotations
from typing import TextIO
import os
import io

import ezdxf
from ezdxf.document import Drawing
from ezdxf.layouts import Modelspace
from ezdxf.lldxf.tagwriter import TagWriter, AbstractTagWriter

__all__ = ["export_file", "export_stream"]


def export_file(doc: Drawing, filename: str | os.PathLike) -> None:
    """Exports the specified DXF R12 document, which should contain content conforming
    to the ASTM-D6673-10 standard, in a special way so that Gerber Technology applications
    can parse it by their low-quality DXF parser.
    """
    fp = io.open(filename, mode="wt", encoding="ascii", errors="dxfreplace")
    export_stream(doc, fp)


def export_stream(doc: Drawing, stream: TextIO) -> None:
    """Exports the specified DXF R12 document into a `stream` object."""

    if doc.dxfversion != ezdxf.const.DXF12:
        raise ezdxf.DXFVersionError("only DXF R12 format is supported")
    tagwriter = TagWriter(stream, write_handles=False, dxfversion=ezdxf.const.DXF12)
    _export_sections(doc, tagwriter)


def _export_sections(doc: Drawing, tagwriter: AbstractTagWriter) -> None:
    _export_header(tagwriter)
    _export_blocks(doc, tagwriter)
    _export_entities(doc.modelspace(), tagwriter)


def _export_header(tagwriter: AbstractTagWriter) -> None:
    # export empty header section
    tagwriter.write_str("  0\nSECTION\n  2\nHEADER\n")
    tagwriter.write_tag2(0, "ENDSEC")


def _export_blocks(doc: Drawing, tagwriter: AbstractTagWriter) -> None:
    # This is the important part:
    #
    # Gerber Technology applications have a bad DXF parser which do not accept DXF
    # files that contain blocks without ASTM-D6673-10 content, such as the standard
    # $MODEL_SPACE and $PAPER_SPACE block definitions.
    #
    # This is annoying but the presence of these blocks is NOT mandatory for
    # the DXF R12 standard.
    #
    tagwriter.write_str("  0\nSECTION\n  2\nBLOCKS\n")
    for block_record in doc.block_records:
        # export only BLOCK definitions, ignore LAYOUT definition blocks
        if block_record.is_block_layout:
            block_record.export_block_definition(tagwriter)
    tagwriter.write_tag2(0, "ENDSEC")


def _export_entities(msp: Modelspace, tagwriter: AbstractTagWriter) -> None:
    tagwriter.write_str("  0\nSECTION\n  2\nENTITIES\n")
    msp.entity_space.export_dxf(tagwriter)
    tagwriter.write_tag2(0, "ENDSEC")
    tagwriter.write_tag2(0, "EOF")



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.addons import meshex
from ezdxf.render import MeshBuilder
from ezdxf.render.forms import cube
from ezdxf.version import __version__


class TestStlLoader:
    def test_empty_file_returns_empty_mesh(self):
        mesh = meshex.stl_loads("")
        assert len(mesh.vertices) == 0

    def test_load_a_single_face(self):
        """This is the minimum required format accuracy the parser needs"""
        mesh = meshex.stl_loads(
            "vertex 0 0 0\nvertex 1 0 0\nvertex 1 1 0\nendloop\n"
        )
        assert mesh.vertices[0] == (0, 0, 0)
        assert mesh.vertices[1] == (1, 0, 0)
        assert mesh.vertices[2] == (1, 1, 0)
        assert mesh.faces[0] == (0, 1, 2)

    def test_parsing_error_too_few_axis(self):
        with pytest.raises(meshex.ParsingError):
            meshex.stl_loads("vertex 0 0")

    def test_parsing_error_invalid_coordinate_format(self):
        with pytest.raises(meshex.ParsingError):
            meshex.stl_loads("vertex 0, 0, 0")

    def test_parsing_error_invalid_floats(self):
        with pytest.raises(meshex.ParsingError):
            meshex.stl_loads("vertex 0 0 z")


OFF_VALID_1 = """OFF
# just a comment
8 6 0
-0.500000 -0.500000 0.500000  #  ignor this
0.500000 -0.500000 0.500000
-0.500000 0.500000 0.500000
0.500000 0.500000 0.500000
-0.500000 0.500000 -0.500000
0.500000 0.500000 -0.500000
-0.500000 -0.500000 -0.500000
0.500000 -0.500000 -0.500000
4 0 1 3 2  # ignore this
4 2 3 5 4
4 4 5 7 6
4 6 7 1 0
4 1 7 5 3
4 6 0 2 4
"""

OFF_VALID_2 = """OFF 8 6 0  # ignore this
-0.500000 -0.500000 0.500000  #  ignore this
0.500000 -0.500000 0.500000
-0.500000 0.500000 0.500000
0.500000 0.500000 0.500000
-0.500000 0.500000 -0.500000
0.500000 0.500000 -0.500000
-0.500000 -0.500000 -0.500000
0.500000 -0.500000 -0.500000
4 0 1 3 2  # ignore this
4 2 3 5 4
4 4 5 7 6
4 6 7 1 0
4 1 7 5 3
4 6 0 2 4


# ignore this
"""


class TestOFFLoader:
    @pytest.mark.parametrize("data", [OFF_VALID_1, OFF_VALID_2])
    def test_valid_data(self, data):
        mesh = meshex.off_loads(data)
        assert len(mesh.vertices) == 8
        assert len(mesh.faces) == 6

    def test_minimal_data(self):
        mesh = meshex.off_loads("3 1 0\n0 0 0\n 0 1 0\n1 1 0\n3 0 1 2")
        assert len(mesh.vertices) == 3
        assert len(mesh.faces) == 1

    @pytest.mark.parametrize(
        "data",
        [
            "",  # no data
            "OFF",  # no data
            "OFF 8 6 0",  # invalid data count
            "OFF 8 6 0\n1 2 3",  # invalid data count
            "OFF 3 1 0\n0 0 0\n 0 0 0\n0 0 Z\n3 0 1 2\n ",  # vertex parsing error
            "OFF 3 1 0\n0 0 0\n 0 0 0\n0 0\n3 0 1 2\n ",  # vertex parsing error
            "OFF 3 1 0\n0 0 0\n 0 0 0\n0 0 0\n3 0 1\n ",  # face parsing error
            "OFF 3 1 0\n0 0 0\n 0 0 0\n0 0 0\n3 0 1 z\n ",  # face parsing error
        ],
    )
    def test_parsing_error_invalid_data_raises_parsing_error(self, data):
        with pytest.raises(meshex.ParsingError):
            meshex.off_loads(data)


class TestOBJLoader:
    def test_empty_file_returns_empty_mesh(self):
        meshes = meshex.obj_loads("")
        assert len(meshes) == 0

    @pytest.mark.parametrize(
        "s",
        [
            "v 0 0 0\nv 1 0 0\nv 1 1 0\nf 1 2 3\n",
            "g\nv 0 0 0\nv 1 0 0\nv 1 1 0\ng\nf 1 2 3\n",
        ],
    )
    def test_load_a_single_face(self, s):
        meshes = meshex.obj_loads(s)  # OBJ is 1-indexed
        mesh = meshes[0]
        assert mesh.vertices[0] == (0, 0, 0)
        assert mesh.vertices[1] == (1, 0, 0)
        assert mesh.vertices[2] == (1, 1, 0)
        assert mesh.faces[0] == (0, 1, 2)

    def test_parsing_error_too_few_axis(self):
        with pytest.raises(meshex.ParsingError):
            meshex.obj_loads("v 0 0")

    def test_parsing_error_invalid_coordinate_format(self):
        with pytest.raises(meshex.ParsingError):
            meshex.obj_loads("v 0, 0, 0")

    def test_parsing_error_invalid_floats(self):
        with pytest.raises(meshex.ParsingError):
            meshex.obj_loads("v 0 0 z")


STL_SINGLE_TRIANGLE = f"""solid STL generated by ezdxf {__version__}
  facet normal 0.0 0.0 1.0
    outer loop
      vertex 0.0 0.0 0.0
      vertex 1.0 0.0 0.0
      vertex 1.0 1.0 0.0
    endloop
  endfacet
endsolid
"""


class TestSTLDumpString:
    def test_single_face(self):
        mesh = MeshBuilder()
        mesh.add_face([(0, 0), (1, 0), (1, 1)])
        result = meshex.stl_dumps(mesh)
        assert result == STL_SINGLE_TRIANGLE

    def test_dump_and_load_cube_ascii(self):
        mesh = meshex.stl_loads(meshex.stl_dumps(cube()))
        assert len(mesh.vertices) == 8
        assert len(mesh.faces) == 12

    def test_dump_and_load_cube_binary(self):
        mesh = meshex.stl_loadb(meshex.stl_dumpb(cube()))
        assert len(mesh.vertices) == 8
        assert len(mesh.faces) == 12


OFF_SINGLE_TRIANGLE = """OFF
3 1 0
0.0 0.0 0.0
1.0 0.0 0.0
1.0 1.0 0.0
3 0 1 2
"""


class TestOFFDumpString:
    def test_single_face(self):
        mesh = MeshBuilder()
        mesh.add_face([(0, 0), (1, 0), (1, 1)])
        result = meshex.off_dumps(mesh)
        assert result == OFF_SINGLE_TRIANGLE

    def test_dump_and_load_cube(self):
        mesh = meshex.off_loads(meshex.off_dumps(cube()))
        assert len(mesh.vertices) == 8
        assert len(mesh.faces) == 6


OBJ_SINGLE_TRIANGLE = f"""# OBJ generated by ezdxf {__version__}
v 0.0 0.0 0.0
v 1.0 0.0 0.0
v 1.0 1.0 0.0
f 1 2 3
"""


class TestOBJDumpString:
    def test_single_face(self):
        mesh = MeshBuilder()
        mesh.add_face([(0, 0), (1, 0), (1, 1)])
        result = meshex.obj_dumps(mesh)
        assert result == OBJ_SINGLE_TRIANGLE

    def test_dump_and_load_cube(self):
        mesh = meshex.obj_loads(meshex.obj_dumps(cube()))[0]
        assert len(mesh.vertices) == 8
        assert len(mesh.faces) == 6


OPENSCAD_SINGLE_TRIANGLE = """polyhedron(points = [
  [0.0, 0.0, 0.0],
  [1.0, 0.0, 0.0],
  [1.0, 1.0, 0.0],
], faces = [
  [0, 1, 2],
], convexity = 10);
"""

OPENSCAD_INVERTED_CUBE = """polyhedron(points = [
  [-0.5, -0.5, -0.5],
  [0.5, -0.5, -0.5],
  [0.5, 0.5, -0.5],
  [-0.5, 0.5, -0.5],
  [-0.5, -0.5, 0.5],
  [0.5, -0.5, 0.5],
  [0.5, 0.5, 0.5],
  [-0.5, 0.5, 0.5],
], faces = [
  [1, 2, 3, 0],
  [5, 6, 2, 1],
  [2, 6, 7, 3],
  [3, 7, 4, 0],
  [4, 5, 1, 0],
  [7, 6, 5, 4],
], convexity = 10);
"""


class TestOpenSCADDumpString:
    def test_single_face(self):
        mesh = MeshBuilder()
        mesh.add_face([(0, 0), (1, 0), (1, 1)])
        result = meshex.scad_dumps(mesh)
        assert result == OPENSCAD_SINGLE_TRIANGLE

    def test_inverted_cube(self):
        """OpenSCAD expects inward pointing normals, make sure to flip normals
        before exporting to OpenSCAD! The import by OpenSCAD works also with
        outwards pointing normals but boolean operations will fail.
        """
        mesh = cube()
        mesh.flip_normals()
        result = meshex.scad_dumps(mesh)
        assert result == OPENSCAD_INVERTED_CUBE


def test_ifc_guid_compression():
    guid_hex = "a1207e36d39f4c629e6544dc9d6e27c9"
    ifc_guid = meshex._guid_compress(guid_hex)
    assert len(ifc_guid) == 22
    assert ifc_guid == "2X87usqvzCOfvbHDoTRYV9"


class TestRecords:
    @pytest.fixture
    def records(self):
        return meshex.Records()

    def test_add_record_returns_record_number(self, records):
        rec = records.add("IFCSHAPEREPRESENTATION(#31,'Body','Brep',#32);")
        assert rec == "#1"
        assert records.prev_num == 0
        assert records.last_num == 1
        assert records.next_num == 2

    def test_add_record_check_expected_record_number(self, records):
        with pytest.raises(ValueError):
            records.add("IFCSHAPEREPRESENTATION(#31,'Body','Brep',#32);", 7)

    def test_update_one_tag(self, records):
        records.add("IFCSHAPEREPRESENTATION($ENTITY$);")
        rec = records.add("IFCENTITY();")
        records.update_all("$ENTITY$", rec)
        assert records.get(1) == "IFCSHAPEREPRESENTATION(#2);"

    def test_update_multiple_tags(self, records):
        records.add("IFCSHAPEREPRESENTATION($ENTITY$,$XREF$);")
        rec1 = records.add("IFCENTITY();")
        rec2 = records.add("IFCXREF();")
        records.update_all("$ENTITY$", rec1)
        records.update_all("$XREF$", rec2)
        assert records.get(1) == "IFCSHAPEREPRESENTATION(#2,#3);"

    def test_dumps(self, records):
        records.add("IFCTAG1();")
        records.add("IFCTAG2();")
        assert records.dumps() == "#1= IFCTAG1();\n#2= IFCTAG2();"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2020-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Union, NamedTuple, Sequence
from typing_extensions import Self
import math


class RGB(NamedTuple):
    """Named tuple representing an RGB color value.

    Attributes:
        r: red channel in range [0, 255]
        g: green channel in range [0, 255]
        b: blue channel in range [0, 255]
    """

    r: int
    g: int
    b: int

    def to_floats(self) -> tuple[float, float, float]:
        """Returns the color value as a tuple of floats in range [0, 1]."""
        return self.r / 255, self.g / 255, self.b / 255

    @classmethod
    def from_floats(cls, rgb: tuple[float, float, float]) -> Self:
        """Returns an :class:`RGB` instance from floats in range [0, 1]."""
        r = max(min(255, round(rgb[0] * 255)), 0)
        g = max(min(255, round(rgb[1] * 255)), 0)
        b = max(min(255, round(rgb[2] * 255)), 0)
        return cls(r, g, b)

    def to_hex(self) -> str:
        """Returns the color value as hex string "#RRGGBB"."""
        return f"#{self.r:02x}{self.g:02x}{self.b:02x}"

    @classmethod
    def from_hex(cls, color: str) -> Self:
        """Returns an :class:`RGB` instance from a hex color string, the `color` string
        is a hex string "RRGGBB" with an optional leading "#", an appended alpha
        channel is ignore.
        """
        hex_string = color.lstrip("#")
        r = int(hex_string[0:2], 16)
        g = int(hex_string[2:4], 16)
        b = int(hex_string[4:6], 16)
        return cls(r, g, b)

    @property
    def luminance(self) -> float:
        """Returns perceived luminance for an RGB color in range [0.0, 1.0]
        from dark to light.
        """
        return luminance(self)


class RGBA(NamedTuple):
    """Named tuple representing an RGBA color value. 
    The default alpha channel is 255 (opaque).

    Attributes:
        r: red channel in range [0, 255]
        g: green channel in range [0, 255]
        b: blue channel in range [0, 255]
        a: alpha channel in range [0, 255], where 0 is transparent and 255 is opaque
    """

    r: int
    g: int
    b: int
    a: int = 255  # default alpha channel is opaque

    def to_floats(self) -> tuple[float, float, float, float]:
        """Returns the color value as a tuple of floats in range [0, 1]."""
        return self.r / 255, self.g / 255, self.b / 255, self.a / 255

    @classmethod
    def from_floats(cls, values: Sequence[float]) -> Self:
        """Returns an :class:`RGBA` instance from floats in range [0, 1].

        The alpha channel is optional. 
        The default alpha channel is 255 (opaque).
        """
        r = max(min(255, round(values[0] * 255)), 0)
        g = max(min(255, round(values[1] * 255)), 0)
        b = max(min(255, round(values[2] * 255)), 0)
        try:
            a = max(min(255, round(values[3] * 255)), 0)
        except IndexError:
            a = 255  # default to opaque == RGB(r, g, b)
        return cls(r, g, b, a)

    def to_hex(self) -> str:
        """Returns the color value as hex string "#RRGGBBAA"."""
        return f"#{self.r:02x}{self.g:02x}{self.b:02x}{self.a:02x}"

    @classmethod
    def from_hex(cls, color: str) -> Self:
        """Returns an :class:`RGBA` instance from a hex color string, the `color` string
        is a hex string "RRGGBBAA" with an optional leading "#". 
        The alpha channel is optional. 
        The default alpha channel is 255 (opaque).
        """
        hex_string = color.lstrip("#")
        r = int(hex_string[0:2], 16)
        g = int(hex_string[2:4], 16)
        b = int(hex_string[4:6], 16)
        a = 255
        if len(hex_string) > 6:
            a = int(hex_string[6:8], 16)
        return cls(r, g, b, a)

    @property
    def luminance(self) -> float:
        """Returns perceived luminance for an RGB color in range [0.0, 1.0]
        from dark to light.
        """
        return luminance(self)


BYBLOCK = 0
BYLAYER = 256
BYOBJECT = 257
RED = 1
YELLOW = 2
GREEN = 3
CYAN = 4
BLUE = 5
MAGENTA = 6
BLACK = 7
WHITE = 7
GRAY = 8
LIGHT_GRAY = 9

# Flags for raw color int values:
# Take color from layer, ignore other bytes.
COLOR_TYPE_BY_LAYER = 0xC0
# Take color from insertion, ignore other bytes
COLOR_TYPE_BY_BLOCK = 0xC1
# RGB value, other bytes are R,G,B.
COLOR_TYPE_RGB = 0xC2
# ACI, AutoCAD color index, other bytes are 0,0,index ???
COLOR_TYPE_ACI = 0xC3

# Found in MLEADER text background color (group code 91) = -939524096
# guess: use window background color
COLOR_TYPE_WINDOW_BG = 0xC8


def decode_raw_color(value: int) -> tuple[int, Union[int, RGB]]:
    """Decode :term:`raw-color` value as tuple(type, Union[aci, (r, g, b)]), the
    true color value is a (r, g, b) tuple.
    """
    data = decode_raw_color_int(value)
    if data[0] == COLOR_TYPE_RGB:
        return COLOR_TYPE_RGB, int2rgb(data[1])
    return data


def decode_raw_color_int(value: int) -> tuple[int, int]:
    """Decode :term:`raw-color` value as tuple(type, int), the true color value
    is a 24-bit int value.
    """
    flags = (value >> 24) & 0xFF
    if flags == COLOR_TYPE_BY_BLOCK:
        return COLOR_TYPE_BY_BLOCK, BYBLOCK
    elif flags == COLOR_TYPE_BY_LAYER:
        return COLOR_TYPE_BY_LAYER, BYLAYER
    elif flags == COLOR_TYPE_ACI:
        return COLOR_TYPE_ACI, value & 0xFF
    elif flags == COLOR_TYPE_RGB:
        return COLOR_TYPE_RGB, value & 0xFFFFFF
    elif flags == COLOR_TYPE_WINDOW_BG:
        return COLOR_TYPE_WINDOW_BG, 0
    else:
        raise ValueError(f"Unknown color type: 0x{flags:02x}")


BY_LAYER_RAW_VALUE = -1073741824  # -(-(0xc0 << 24) & 0xffffffff)
BY_BLOCK_RAW_VALUE = -1056964608  # -(-(0xc1 << 24) & 0xffffffff)
WINDOW_BG_RAW_VALUE = -939524096


def encode_raw_color(value: Union[int, RGB]) -> int:
    """Encode :term:`true-color` value or :ref:`ACI` color value into a :term:
    `raw color` value.
    """
    if isinstance(value, int):
        if value == BYBLOCK:
            return BY_BLOCK_RAW_VALUE
        elif value == BYLAYER:
            return BY_LAYER_RAW_VALUE
        elif 0 < value < 256:
            return -(-(COLOR_TYPE_ACI << 24) & 0xFFFFFFFF) | value
        else:  # BYOBJECT (257) -> resolve to object color
            raise ValueError(f"Invalid color index: {value}")
    else:
        return -(-((COLOR_TYPE_RGB << 24) + rgb2int(value)) & 0xFFFFFFFF)


TRANSPARENCY_BYBLOCK = 0x1000000
OPAQUE = 0x20000FF
TRANSPARENCY_10 = 0x20000E5
TRANSPARENCY_20 = 0x20000CC
TRANSPARENCY_30 = 0x20000B2
TRANSPARENCY_40 = 0x2000099
TRANSPARENCY_50 = 0x200007F
TRANSPARENCY_60 = 0x2000066
TRANSPARENCY_70 = 0x200004C
TRANSPARENCY_80 = 0x2000032
TRANSPARENCY_90 = 0x2000019


def float2transparency(value: float) -> int:
    """
    Returns DXF transparency value as integer in the range from 0 to 255,
    where 0 is 100% transparent and 255 is opaque.

    Args:
        value: transparency value as float in the range from 0 to 1, where 0 is
            opaque and 1 is 100% transparent.

    """
    return int((1.0 - float(value)) * 255) | 0x02000000


def transparency2float(value: int) -> float:
    """
    Returns transparency value as float from 0 to 1, 0 for no transparency
    (opaque) and 1 for 100% transparency.

    Args:
        value: DXF integer transparency value, 0 for 100% transparency and 255
            for opaque

    """
    # Transparency value 0x020000TT 0 = fully transparent / 255 = opaque
    # 255 -> 0.0
    # 0 -> 1.0
    return 1.0 - float(int(value) & 0xFF) / 255.0


def int2rgb(value: int) -> RGB:
    """Split RGB integer `value` into (r, g, b) tuple."""
    return RGB(
        (value >> 16) & 0xFF,  # red
        (value >> 8) & 0xFF,  # green
        value & 0xFF,  # blue
    )


def rgb2int(rgb: RGB | tuple[int, int, int]) -> int:
    """Combined integer value from (r, g, b) tuple."""
    r, g, b = rgb
    return ((int(r) & 0xFF) << 16) | ((int(g) & 0xFF) << 8) | (int(b) & 0xFF)


def aci2rgb(index: int) -> RGB:
    """Convert :ref:`ACI` into (r, g, b) tuple, based on default AutoCAD
    colors.
    """
    if index < 1:
        raise IndexError(index)
    return int2rgb(DXF_DEFAULT_COLORS[index])


def luminance(color: Sequence[float]) -> float:
    """Returns perceived luminance for an RGB color in the range [0.0, 1.0]
    from dark to light.
    """
    r = float(color[0]) / 255
    g = float(color[1]) / 255
    b = float(color[2]) / 255
    return round(math.sqrt(0.299 * r * r + 0.587 * g * g + 0.114 * b * b), 3)


# color codes are 1-indexed so an additional entry was put in the 0th position
# different plot styles may choose different colors for the same code
# from ftp://ftp.ecn.purdue.edu/jshan/86/help/html/import_export/dxf_colortable.htm
# alternative color tables can be found at:
#  - https://www.temblast.com/songview/color3.htm
#  - https://gohtx.com/acadcolors.php

# modelspace color palette for AutoCAD 2020
#  - DXF_DEFAULT_COLORS values change depending on Uniform background
#    color setting in options, these values correspond the default
#    2D model space Uniform background RGB color for AutoCAD 2020: 33,40,48
DXF_DEFAULT_COLORS = [
    0x000000,  # dummy value for index [0]
    0xFF0000,
    0xFFFF00,
    0x00FF00,
    0x00FFFF,
    0x0000FF,
    0xFF00FF,
    0xFFFFFF,
    0x808080,
    0xC0C0C0,
    0xFF0000,
    0xFF7F7F,
    0xA50000,
    0xA55252,
    0x7F0000,
    0x7F3F3F,
    0x4C0000,
    0x4C2626,
    0x260000,
    0x261313,
    0xFF3F00,
    0xFF9F7F,
    0xA52900,
    0xA56752,
    0x7F1F00,
    0x7F4F3F,
    0x4C1300,
    0x4C2F26,
    0x260900,
    0x261713,
    0xFF7F00,
    0xFFBF7F,
    0xA55200,
    0xA57C52,
    0x7F3F00,
    0x7F5F3F,
    0x4C2600,
    0x4C3926,
    0x261300,
    0x261C13,
    0xFFBF00,
    0xFFDF7F,
    0xA57C00,
    0xA59152,
    0x7F5F00,
    0x7F6F3F,
    0x4C3900,
    0x4C4226,
    0x261C00,
    0x262113,
    0xFFFF00,
    0xFFFF7F,
    0xA5A500,
    0xA5A552,
    0x7F7F00,
    0x7F7F3F,
    0x4C4C00,
    0x4C4C26,
    0x262600,
    0x262613,
    0xBFFF00,
    0xDFFF7F,
    0x7CA500,
    0x91A552,
    0x5F7F00,
    0x6F7F3F,
    0x394C00,
    0x424C26,
    0x1C2600,
    0x212613,
    0x7FFF00,
    0xBFFF7F,
    0x52A500,
    0x7CA552,
    0x3F7F00,
    0x5F7F3F,
    0x264C00,
    0x394C26,
    0x132600,
    0x1C2613,
    0x3FFF00,
    0x9FFF7F,
    0x29A500,
    0x67A552,
    0x1F7F00,
    0x4F7F3F,
    0x134C00,
    0x2F4C26,
    0x092600,
    0x172613,
    0x00FF00,
    0x7FFF7F,
    0x00A500,
    0x52A552,
    0x007F00,
    0x3F7F3F,
    0x004C00,
    0x264C26,
    0x002600,
    0x132613,
    0x00FF3F,
    0x7FFF9F,
    0x00A529,
    0x52A567,
    0x007F1F,
    0x3F7F4F,
    0x004C13,
    0x264C2F,
    0x002609,
    0x135817,
    0x00FF7F,
    0x7FFFBF,
    0x00A552,
    0x52A57C,
    0x007F3F,
    0x3F7F5F,
    0x004C26,
    0x264C39,
    0x002613,
    0x13581C,
    0x00FFBF,
    0x7FFFDF,
    0x00A57C,
    0x52A591,
    0x007F5F,
    0x3F7F6F,
    0x004C39,
    0x264C42,
    0x00261C,
    0x135858,
    0x00FFFF,
    0x7FFFFF,
    0x00A5A5,
    0x52A5A5,
    0x007F7F,
    0x3F7F7F,
    0x004C4C,
    0x264C4C,
    0x002626,
    0x135858,
    0x00BFFF,
    0x7FDFFF,
    0x007CA5,
    0x5291A5,
    0x005F7F,
    0x3F6F7F,
    0x00394C,
    0x26427E,
    0x001C26,
    0x135858,
    0x007FFF,
    0x7FBFFF,
    0x0052A5,
    0x527CA5,
    0x003F7F,
    0x3F5F7F,
    0x00264C,
    0x26397E,
    0x001326,
    0x131C58,
    0x003FFF,
    0x7F9FFF,
    0x0029A5,
    0x5267A5,
    0x001F7F,
    0x3F4F7F,
    0x00134C,
    0x262F7E,
    0x000926,
    0x131758,
    0x0000FF,
    0x7F7FFF,
    0x0000A5,
    0x5252A5,
    0x00007F,
    0x3F3F7F,
    0x00004C,
    0x26267E,
    0x000026,
    0x131358,
    0x3F00FF,
    0x9F7FFF,
    0x2900A5,
    0x6752A5,
    0x1F007F,
    0x4F3F7F,
    0x13004C,
    0x2F267E,
    0x090026,
    0x171358,
    0x7F00FF,
    0xBF7FFF,
    0x5200A5,
    0x7C52A5,
    0x3F007F,
    0x5F3F7F,
    0x26004C,
    0x39267E,
    0x130026,
    0x1C1358,
    0xBF00FF,
    0xDF7FFF,
    0x7C00A5,
    0x9152A5,
    0x5F007F,
    0x6F3F7F,
    0x39004C,
    0x42264C,
    0x1C0026,
    0x581358,
    0xFF00FF,
    0xFF7FFF,
    0xA500A5,
    0xA552A5,
    0x7F007F,
    0x7F3F7F,
    0x4C004C,
    0x4C264C,
    0x260026,
    0x581358,
    0xFF00BF,
    0xFF7FDF,
    0xA5007C,
    0xA55291,
    0x7F005F,
    0x7F3F6F,
    0x4C0039,
    0x4C2642,
    0x26001C,
    0x581358,
    0xFF007F,
    0xFF7FBF,
    0xA50052,
    0xA5527C,
    0x7F003F,
    0x7F3F5F,
    0x4C0026,
    0x4C2639,
    0x260013,
    0x58131C,
    0xFF003F,
    0xFF7F9F,
    0xA50029,
    0xA55267,
    0x7F001F,
    0x7F3F4F,
    0x4C0013,
    0x4C262F,
    0x260009,
    0x581317,
    0x000000,
    0x656565,
    0x666666,
    0x999999,
    0xCCCCCC,
    0xFFFFFF,
]


# paperspace color palette for AutoCAD 2020
#  - DXF_DEFAULT_PAPERSPACE_COLORS values change depending on Uniform background
#    color setting in options, these values correspond the default
#    Sheet/layout Uniform background color for AutoCAD 2020: White (RGB 255,255,255)
DXF_DEFAULT_PAPERSPACE_COLORS = [
    0x000000,  # dummy value for index [0]
    0xFF0000,
    0xFFFF00,
    0x00FF00,
    0x00FFFF,
    0x0000FF,
    0xFF00FF,
    0x000000,
    0x808080,
    0xC0C0C0,
    0xFF0000,
    0xFF7F7F,
    0xA50000,
    0xA55252,
    0x7F0000,
    0x7F3F3F,
    0x4C0000,
    0x4C2626,
    0x260000,
    0x261313,
    0xFF3F00,
    0xFF9F7F,
    0xA52900,
    0xA56752,
    0x7F1F00,
    0x7F4F3F,
    0x4C1300,
    0x4C2F26,
    0x260900,
    0x261713,
    0xFF7F00,
    0xFFBF7F,
    0xA55200,
    0xA57C52,
    0x7F3F00,
    0x7F5F3F,
    0x4C2600,
    0x4C3926,
    0x261300,
    0x261C13,
    0xFFBF00,
    0xFFDF7F,
    0xA57C00,
    0xA59152,
    0x7F5F00,
    0x7F6F3F,
    0x4C3900,
    0x4C4226,
    0x261C00,
    0x262113,
    0xFFFF00,
    0xFFFF7F,
    0xA5A500,
    0xA5A552,
    0x7F7F00,
    0x7F7F3F,
    0x4C4C00,
    0x4C4C26,
    0x262600,
    0x262613,
    0xBFFF00,
    0xDFFF7F,
    0x7CA500,
    0x91A552,
    0x5F7F00,
    0x6F7F3F,
    0x394C00,
    0x424C26,
    0x1C2600,
    0x212613,
    0x7FFF00,
    0xBFFF7F,
    0x52A500,
    0x7CA552,
    0x3F7F00,
    0x5F7F3F,
    0x264C00,
    0x394C26,
    0x132600,
    0x1C2613,
    0x3FFF00,
    0x9FFF7F,
    0x29A500,
    0x67A552,
    0x1F7F00,
    0x4F7F3F,
    0x134C00,
    0x2F4C26,
    0x092600,
    0x172613,
    0x00FF00,
    0x7FFF7F,
    0x00A500,
    0x52A552,
    0x007F00,
    0x3F7F3F,
    0x004C00,
    0x264C26,
    0x002600,
    0x132613,
    0x00FF3F,
    0x7FFF9F,
    0x00A529,
    0x52A567,
    0x007F1F,
    0x3F7F4F,
    0x004C13,
    0x264C2F,
    0x002609,
    0x132617,
    0x00FF7F,
    0x7FFFBF,
    0x00A552,
    0x52A57C,
    0x007F3F,
    0x3F7F5F,
    0x004C26,
    0x264C39,
    0x002613,
    0x13261C,
    0x00FFBF,
    0x7FFFDF,
    0x00A57C,
    0x52A591,
    0x007F5F,
    0x3F7F6F,
    0x004C39,
    0x264C42,
    0x00261C,
    0x132621,
    0x00FFFF,
    0x7FFFFF,
    0x00A5A5,
    0x52A5A5,
    0x007F7F,
    0x3F7F7F,
    0x004C4C,
    0x264C4C,
    0x002626,
    0x132626,
    0x00BFFF,
    0x7FDFFF,
    0x007CA5,
    0x5291A5,
    0x005F7F,
    0x3F6F7F,
    0x00394C,
    0x26424C,
    0x001C26,
    0x132126,
    0x007FFF,
    0x7FBFFF,
    0x0052A5,
    0x527CA5,
    0x003F7F,
    0x3F5F7F,
    0x00264C,
    0x26394C,
    0x001326,
    0x131C26,
    0x003FFF,
    0x7F9FFF,
    0x0029A5,
    0x5267A5,
    0x001F7F,
    0x3F4F7F,
    0x00134C,
    0x262F4C,
    0x000926,
    0x131726,
    0x0000FF,
    0x7F7FFF,
    0x0000A5,
    0x5252A5,
    0x00007F,
    0x3F3F7F,
    0x00004C,
    0x26264C,
    0x000026,
    0x131326,
    0x3F00FF,
    0x9F7FFF,
    0x2900A5,
    0x6752A5,
    0x1F007F,
    0x4F3F7F,
    0x13004C,
    0x2F264C,
    0x090026,
    0x171326,
    0x7F00FF,
    0xBF7FFF,
    0x5200A5,
    0x7C52A5,
    0x3F007F,
    0x5F3F7F,
    0x26004C,
    0x39264C,
    0x130026,
    0x1C1326,
    0xBF00FF,
    0xDF7FFF,
    0x7C00A5,
    0x9152A5,
    0x5F007F,
    0x6F3F7F,
    0x39004C,
    0x42264C,
    0x1C0026,
    0x211326,
    0xFF00FF,
    0xFF7FFF,
    0xA500A5,
    0xA552A5,
    0x7F007F,
    0x7F3F7F,
    0x4C004C,
    0x4C264C,
    0x260026,
    0x261326,
    0xFF00BF,
    0xFF7FDF,
    0xA5007C,
    0xA55291,
    0x7F005F,
    0x7F3F6F,
    0x4C0039,
    0x4C2642,
    0x26001C,
    0x261321,
    0xFF007F,
    0xFF7FBF,
    0xA50052,
    0xA5527C,
    0x7F003F,
    0x7F3F5F,
    0x4C0026,
    0x4C2639,
    0x260013,
    0x26131C,
    0xFF003F,
    0xFF7F9F,
    0xA50029,
    0xA55267,
    0x7F001F,
    0x7F3F4F,
    0x4C0013,
    0x4C262F,
    0x260009,
    0x261317,
    0x000000,
    0x2D2D2D,
    0x5B5B5B,
    0x898989,
    0xB7B7B7,
    0xB3B3B3,
]



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import math

from ezdxf.math import ellipse_param_span, arc_angle_span_deg, arc_angle_span_rad


class TestArcParamSpan:
    def span(self, s, e):
        return arc_angle_span_deg(s, e)

    @pytest.mark.parametrize(
        "start, end",
        [
            (0, 0),
            (90, 90),
            (-90, -90),
            (180, 180),
            (-180, -180),
            (270, 270),
            (-270, -270),
            (360, 360),
            (-360, -360),
            (720, 720),
            (-720, -720),
        ],
    )
    def test_no_curve(self, start, end):
        assert self.span(start, end) == pytest.approx(0)

    @pytest.mark.parametrize(
        "start, end",
        [
            # full circles:
            # Normalized start- and end angles are equal, but input values are
            # different:
            (0, 360),
            (360, 0),
            (-360, 0),
            (0, -360),
            (90, 450),
            (180, 540),
            (180, -180),
            (-180, 180),
            (90, -270),
            (-90, 270),
        ],
    )
    def test_closed_curve(self, start, end):
        assert self.span(start, end) == pytest.approx(360.0)

    @pytest.mark.parametrize(
        "start, end, expected",
        [
            (0, 90, 90),
            (0, -90, 270),
            (0, 180, 180),
            (0, -180, 180),
            (180, 360, 180),
            (-180, -360, 180),
            (-90, 360, 90),
            (90, -360, 270),
            (-90, -360, 90),
            (90, 360, 270),
            (360, 90, 90),  # start angle 360 is 0
            (360, -90, 270),  # start angle 360 is 0
            (-360, 90, 90),  # start angle -360 is 0
            (-360, -90, 270),  # start angle -360 is 0
            (30, -30, 300),  # crossing 0 deg
            (-30, 30, 60),  # crossing 0 deg
            (90, -90, 180),
            (-90, 90, 180),
            (360, 400, 40),
            (400, 360, 320),
        ],
    )
    def test_partial_arc(self, start, end, expected):
        assert self.span(start, end) == pytest.approx(expected)


class TestEllipseParamSpan(TestArcParamSpan):
    def span(self, s, e):
        return math.degrees(
            ellipse_param_span(math.radians(s), math.radians(e))
        )


class TestArcParamSpanRadians(TestArcParamSpan):
    def span(self, s, e):
        return math.degrees(
            arc_angle_span_rad(math.radians(s), math.radians(e))
        )


PI2 = math.pi / 2.0


class TestParamSpan:
    # Preserved old param span test.
    # This tests work without degrees-radians-degrees conversion
    @pytest.mark.parametrize(
        "start, end",
        [
            (0, 0),
            (math.pi, math.pi),
            (math.tau, math.tau),
            (0, 0),
            (-math.pi, -math.pi),
            (-math.tau, -math.tau),
        ],
    )
    def test_no_ellipse(self, start, end):
        assert ellipse_param_span(start, end) == 0.0

    @pytest.mark.parametrize(
        "start, end",
        [
            (0, math.tau),
            (math.tau, 0),
            (math.pi, -math.pi),
            (0, -math.tau),
            (-math.tau, 0),
            (-math.pi, math.pi),
        ],
    )
    def test_full_ellipse(self, start, end):
        assert ellipse_param_span(start, end) == pytest.approx(math.tau)

    @pytest.mark.parametrize(
        "start, end, expected",
        [
            (0, PI2, PI2),
            (PI2, 0, math.pi * 1.5),
            (PI2, math.pi, PI2),
            (PI2, -PI2, math.pi),
            (math.pi, 0, math.pi),
            (0, math.pi, math.pi),
            (0, -PI2, math.pi * 1.5),
            (-PI2, 0, PI2),
            (-PI2, -math.pi, math.pi * 1.5),
            (-PI2, PI2, math.pi),
            (-math.pi, 0, math.pi),
            (0, -math.pi, math.pi),
        ],
    )
    def test_elliptic_arc(self, start, end, expected):
        assert ellipse_param_span(start, end) == pytest.approx(expected)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Any, no_type_check
from fontTools.pens.basePen import BasePen
from fontTools.ttLib import TTFont

from ezdxf.math import Matrix44, UVec, BoundingBox2d
from ezdxf.path import Path
from .font_manager import FontManager, UnsupportedFont
from .font_measurements import FontMeasurements
from .glyphs import GlyphPath, Glyphs

UNICODE_WHITE_SQUARE = 9633  # U+25A1
UNICODE_REPLACEMENT_CHAR = 65533  # U+FFFD

font_manager = FontManager()


class PathPen(BasePen):
    def __init__(self, glyph_set) -> None:
        super().__init__(glyph_set)
        self._path = Path()

    @property
    def path(self) -> GlyphPath:
        return GlyphPath(self._path)

    def _moveTo(self, pt: UVec) -> None:
        self._path.move_to(pt)

    def _lineTo(self, pt: UVec) -> None:
        self._path.line_to(pt)

    def _curveToOne(self, pt1: UVec, pt2: UVec, pt3: UVec) -> None:
        self._path.curve4_to(pt3, pt1, pt2)

    def _qCurveToOne(self, pt1: UVec, pt2: UVec) -> None:
        self._path.curve3_to(pt2, pt1)

    def _closePath(self) -> None:
        self._path.close_sub_path()


class NoKerning:
    def get(self, c0: str, c1: str) -> float:
        return 0.0


class KerningTable(NoKerning):
    __slots__ = ("_cmap", "_kern_table")

    def __init__(self, font: TTFont, cmap, fmt: int = 0):
        self._cmap = cmap
        self._kern_table = font["kern"].getkern(fmt)

    def get(self, c0: str, c1: str) -> float:
        try:
            return self._kern_table[(self._cmap[ord(c0)], self._cmap[ord(c1)])]
        except (KeyError, TypeError):
            return 0.0


def get_fontname(font: TTFont) -> str:
    names = font["name"].names
    for record in names:
        if record.nameID == 1:
            return record.string.decode(record.getEncoding())
    return "unknown"


class TTFontRenderer(Glyphs):
    def __init__(self, font: TTFont, kerning=False):
        self._glyph_path_cache: dict[str, GlyphPath] = dict()
        self._generic_glyph_cache: dict[str, Any] = dict()
        self._glyph_width_cache: dict[str, float] = dict()
        self.font = font
        self.cmap = self.font.getBestCmap()
        if self.cmap is None:
            raise UnsupportedFont(f"font '{self.font_name}' has no character map.")
        self.glyph_set = self.font.getGlyphSet()
        self.kerning = NoKerning()
        if kerning:
            try:
                self.kerning = KerningTable(self.font, self.cmap)
            except KeyError:  # kerning table does not exist
                pass
        self.undefined_generic_glyph = self.glyph_set[".notdef"]
        self.font_measurements = self._get_font_measurements()
        self.space_width = self.detect_space_width()

    @property
    def font_name(self) -> str:
        return get_fontname(self.font)

    @no_type_check
    def _get_font_measurements(self) -> FontMeasurements:
        bbox = BoundingBox2d(self.get_glyph_path("x").control_vertices())
        baseline = bbox.extmin.y
        x_height = bbox.extmax.y - baseline
        bbox = BoundingBox2d(self.get_glyph_path("A").control_vertices())
        cap_height = bbox.extmax.y - baseline
        bbox = BoundingBox2d(self.get_glyph_path("p").control_vertices())
        descender_height = baseline - bbox.extmin.y
        return FontMeasurements(
            baseline=baseline,
            cap_height=cap_height,
            x_height=x_height,
            descender_height=descender_height,
        )

    def get_scaling_factor(self, cap_height: float) -> float:
        return 1.0 / self.font_measurements.cap_height * cap_height

    def get_generic_glyph(self, char: str):
        try:
            return self._generic_glyph_cache[char]
        except KeyError:
            pass
        try:
            generic_glyph = self.glyph_set[self.cmap[ord(char)]]
        except KeyError:
            generic_glyph = self.undefined_generic_glyph
        self._generic_glyph_cache[char] = generic_glyph
        return generic_glyph

    def get_glyph_path(self, char: str) -> GlyphPath:
        """Returns the raw glyph path, without any scaling applied."""
        try:
            return self._glyph_path_cache[char].clone()
        except KeyError:
            pass
        pen = PathPen(self.glyph_set)
        self.get_generic_glyph(char).draw(pen)
        glyph_path = pen.path
        self._glyph_path_cache[char] = glyph_path
        return glyph_path.clone()

    def get_glyph_width(self, char: str) -> float:
        """Returns the raw glyph width, without any scaling applied."""
        try:
            return self._glyph_width_cache[char]
        except KeyError:
            pass
        width = 0.0
        try:
            width = self.get_generic_glyph(char).width
        except KeyError:
            pass
        self._glyph_width_cache[char] = width
        return width

    def get_text_glyph_paths(
        self, s: str, cap_height: float = 1.0, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        """Returns the glyph paths of string `s` as a list, scaled to cap height."""
        glyph_paths: list[GlyphPath] = []
        x_offset: float = 0
        requires_kerning = isinstance(self.kerning, KerningTable)
        resize_factor = self.get_scaling_factor(cap_height)
        y_factor = resize_factor
        x_factor = resize_factor * width_factor
        # set scaling factor:
        m = Matrix44.scale(x_factor, y_factor, 1.0)
        # set vertical offset:
        m[3, 1] = -self.font_measurements.baseline * y_factor
        prev_char = ""

        for char in s:
            if requires_kerning:
                x_offset += self.kerning.get(prev_char, char) * x_factor
            # set horizontal offset:
            m[3, 0] = x_offset
            glyph_path = self.get_glyph_path(char)
            glyph_path.transform_inplace(m)
            if len(glyph_path):
                glyph_paths.append(glyph_path)
            x_offset += self.get_glyph_width(char) * x_factor
            prev_char = char
        return glyph_paths

    def detect_space_width(self) -> float:
        """Returns the space width for the raw (unscaled) font."""
        return self.get_glyph_width(" ")

    def _get_text_length_with_kerning(self, s: str, cap_height: float = 1.0) -> float:
        length = 0.0
        c0 = ""
        kern = self.kerning.get
        width = self.get_glyph_width
        for c1 in s:
            length += kern(c0, c1) + width(c1)
            c0 = c1
        return length * self.get_scaling_factor(cap_height)

    def get_text_length(
        self, s: str, cap_height: float = 1.0, width_factor: float = 1.0
    ) -> float:
        if isinstance(self.kerning, KerningTable):
            return self._get_text_length_with_kerning(s, cap_height) * width_factor
        width = self.get_glyph_width
        return (
            sum(width(c) for c in s)
            * self.get_scaling_factor(cap_height)
            * width_factor
        )



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.entities.copy import (
    CopyStrategy,
    CopySettings,
    default_copy,
    CopyNotSupported,
)
from ezdxf.entities import Line, DXFObject
from ezdxf.layouts import Modelspace


class UnknownEntity(DXFObject):
    def copy(self, copy_strategy=default_copy):
        raise CopyNotSupported()


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


@pytest.fixture(scope="module")
def msp(doc) -> Modelspace:
    return doc.modelspace()


def test_basic_copy(msp):
    line = msp.add_line(
        (0, 0),
        (10, 0),
        dxfattribs={
            "layer": "MyLayer",
        },
    )
    strategy = CopyStrategy(CopySettings())
    line2 = strategy.copy(line)

    assert line2 is not None
    assert line2 is not line

    assert line2.dxf.start == (0, 0)
    assert line2.dxf.end == (10, 0)
    assert line2.dxf.layer == "MyLayer"


def test_ignore_errors_in_linked_entities(msp: Modelspace):
    line = msp.add_line((0, 0), (10, 0))
    xdict = line.new_extension_dict()
    dictionary = xdict.dictionary
    dictionary["MyKey"] = UnknownEntity()
    clone = line.copy(default_copy)

    xdict = clone.get_extension_dict()
    keys = list(xdict.keys())
    assert len(keys) == 0, "unsupported entry was not copyied"


def test_raise_copy_not_supported_exception(msp: Modelspace):
    line = msp.add_line((0, 0), (10, 0))
    xdict = line.new_extension_dict()
    dictionary = xdict.dictionary
    dictionary["MyKey"] = UnknownEntity()
    copy_strategy = CopyStrategy(
        CopySettings(ignore_copy_errors_in_linked_entities=False)
    )
    with pytest.raises(CopyNotSupported):
        line.copy(copy_strategy)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.enums import MTextParagraphAlignment
from ezdxf.tools.text import MTextContext, ParagraphProperties


def test_underline():
    p = MTextContext()
    assert p.underline is False
    p.underline = True
    assert p.underline is True


def test_strike():
    p = MTextContext()
    assert p.strike_through is False
    p.strike_through = True
    assert p.strike_through is True


def test_overstrike():
    p = MTextContext()
    assert p.overline is False
    p.overline = True
    assert p.overline is True


def test_copy():
    p = MTextContext()
    p.underline = True
    p2 = p.copy()
    p2.underline = False
    assert p != p2


def test_equality():
    p = MTextContext()
    p.underline = True
    p2 = MTextContext()
    p2.underline = True
    assert p == p2


def test_set_aci():
    p = MTextContext()
    p.rgb = (0, 1, 2)
    p.aci = 7
    assert p.aci == 7
    assert p.rgb is None, "should reset rgb value"


class TestParagraphPropertiesToString:
    def test_default_properties(self):
        assert ParagraphProperties().tostring() == ""

    def test_indent_first_line(self):
        assert ParagraphProperties(indent=2).tostring() == "\\pxi2;"

    def test_indent_paragraph_left(self):
        assert ParagraphProperties(left=3).tostring() == "\\pxl3;"

    def test_indent_paragraph_right(self):
        assert ParagraphProperties(right=4).tostring() == "\\pxr4;"

    def test_center_alignment_without_indentation(self):
        assert (
            ParagraphProperties(align=MTextParagraphAlignment.CENTER).tostring()
            == "\\pxqc;"
        )

    def test_center_alignment_with_indentation(self):
        # always a "," after indentations
        assert (
            ParagraphProperties(
                indent=2.5, align=MTextParagraphAlignment.CENTER
            ).tostring()
            == "\\pxi2.5,qc;"
        )

    def test_one_tab_stop(self):
        p = ParagraphProperties(tab_stops=(1,))
        assert p.tostring() == "\\pxt1;"

    def test_multiple_tab_stops(self):
        p = ParagraphProperties(tab_stops=(1, 2, 3))
        assert p.tostring() == "\\pxt1,2,3;"

    def test_different_kinds_of_tab_stops(self):
        # tab stops without prefix or numbers are left adjusted
        # tab stops, e.g 2 or '2'
        # prefix 'c' defines a center adjusted tab stop e.g. 'c3.5'
        # prefix 'r' defines a right adjusted tab stop e.g. 'r2.7'
        p = ParagraphProperties(tab_stops=(1, "c2", "r3.7"))
        assert p.tostring() == "\\pxt1,c2,r3.7;"

    def test_indentation_and_multiple_tab_stops(self):
        p = ParagraphProperties(indent=1, tab_stops=(1, 2, 3))
        # always a "," after indentations
        assert p.tostring() == "\\pxi1,t1,2,3;"

    def test_justified_alignment_and_multiple_tab_stops(self):
        p = ParagraphProperties(
            align=MTextParagraphAlignment.JUSTIFIED, tab_stops=(1, 2, 3)
        )
        assert p.tostring() == "\\pxqj,t1,2,3;"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2010-2020, Manfred Moitzi
# License: MIT License
import pytest
import math
from ezdxf.math import ConstructionRay, ParallelRaysError, Vec3

HALF_PI = math.pi / 2.0


class TestConstructionRay:
    def test_init_with_angle(self):
        point = (10, 10)
        ray = ConstructionRay(point, angle=0)
        ray_normal = ray.orthogonal(point)
        assert ray_normal._is_vertical is True
        ray = ConstructionRay(point, angle=-HALF_PI)
        assert ray._is_horizontal is False
        assert ray._is_vertical is True

    def test_Ray2D_get_x_y(self):
        ray1 = ConstructionRay((10, 1), (20, 10))
        y = ray1.yof(15)
        assert math.isclose(y, 5.5)
        assert math.isclose(ray1.xof(y), 15.0)

    def test_ray2d_intersect(self):
        ray1 = ConstructionRay((10, 1), (20, 10))
        ray2 = ConstructionRay((17, -7), (-10, 3))

        point = ray1.intersect(ray2)
        assert point.isclose(Vec3(5.7434, -2.8309), abs_tol=1e-4)
        assert ray1.is_parallel(ray2) is False

    def test_ray2d_parallel(self):
        ray1 = ConstructionRay((17, -8), (-10, 2))
        ray2 = ConstructionRay((-10, 3), (17, -7))
        ray3 = ConstructionRay((-10, 4), (17, -6))
        assert ray2.is_parallel(ray3) is True
        assert ray1.is_parallel(ray3) is True
        with pytest.raises(ParallelRaysError):
            _ = ray2.intersect(ray3)

    def test_ray2d_intersect_with_vertical(self):
        ray1 = ConstructionRay((10, 1), (10, -7))
        ray2 = ConstructionRay((-10, 3), (17, -7))
        point = ray1.intersect(ray2)
        assert point.x == 10
        assert point.isclose(Vec3(10.0, -4.4074), abs_tol=1e-4)
        with pytest.raises(ArithmeticError):
            _ = ray1.yof(1)

    def test_ray2d_intersect_with_horizontal(self):
        ray1 = ConstructionRay((-10, 10), (10, 10))
        ray2 = ConstructionRay((-10, 20), (10, 0))
        point = ray1.intersect(ray2)
        assert point.y == 10
        assert point.isclose(Vec3(0.0, 10.0), abs_tol=1e-4)

    def test_ray2d_intersect_with_vertical_and_horizontal(self):
        ray1 = ConstructionRay((-10, 10), (10, 10))
        ray2 = ConstructionRay((5, 0), (5, 20))
        point = ray1.intersect(ray2)
        assert point.y == 10
        assert point.x == 5
        assert point.isclose(Vec3(5.0, 10.0), abs_tol=1e-4)

    def test_ray2d_parallel_vertical(self):
        ray1 = ConstructionRay((10, 1), (10, -7))
        ray2 = ConstructionRay((11, 0), angle=HALF_PI)
        ray3 = ConstructionRay((12, -10), (12, 7))
        ray4 = ConstructionRay((0, 0), (1, 1))
        ray5 = ConstructionRay((0, 0), angle=0)
        with pytest.raises(ParallelRaysError):
            _ = ray1.intersect(ray3)
        assert ray1.is_parallel(ray3) is True
        assert ray1.is_parallel(ray2) is True
        assert ray2.is_parallel(ray2) is True
        assert ray1.is_parallel(ray4) is False
        assert ray2.is_parallel(ray4) is False
        assert ray3.is_parallel(ray4) is False
        assert ray1.is_parallel(ray5) is False
        assert ray2.is_parallel(ray5) is False
        assert ray3.is_parallel(ray5) is False
        # vertical rays can't calc a y-value
        with pytest.raises(ArithmeticError):
            _ = ray1.yof(-1.0)

    def test_ray2d_normal_vertical(self):
        ray = ConstructionRay((10, 1), (10, -7))  # vertical line
        ortho = ray.orthogonal((3, 3))
        point = ray.intersect(ortho)
        assert point.isclose(Vec3(10, 3))

    def test_ray2d_normal(self):
        ray = ConstructionRay((-10, 3), (17, -7))
        ortho = ray.orthogonal((3, 3))
        point = ray.intersect(ortho)
        assert point.isclose(Vec3(1.4318, -1.234), abs_tol=1e-4)

    def test_ray2d_normal_horizontal(self):
        ray = ConstructionRay((10, 10), (20, 10))  # horizontal line
        ortho = ray.orthogonal((3, 3))
        point = ray.intersect(ortho)
        assert point.isclose(Vec3(3, 10))

    def test_ray2d_angle(self):
        ray = ConstructionRay((10, 10), angle=HALF_PI)
        assert ray._is_vertical is True
        ray = ConstructionRay((10, 10), angle=0)
        assert ray._is_horizontal is True
        ray = ConstructionRay((10, 10), angle=math.pi / 4)
        assert math.isclose(ray._slope, 1.0)

    def test_bisectrix(self):
        ray1 = ConstructionRay((10, 10), angle=math.pi / 3)
        ray2 = ConstructionRay((3, -5), angle=math.pi / 2)
        ray3 = ConstructionRay((1, 1), angle=math.pi / 3)
        a = ray1.bisectrix(ray2)
        assert math.isclose(a._angle, 1.309, abs_tol=1e-4)
        assert math.isclose(a.yof(7), 12.80385, abs_tol=1e-4)
        with pytest.raises(ParallelRaysError):
            _ = ray1.bisectrix(ray3)

    def test_two_close_horizontal_rays(self):
        p1 = (39340.75302672016, 32489.73349764998)
        p2 = (39037.75302672119, 32489.73349764978)
        p3 = (38490.75302672015, 32489.73349764997)

        ray1 = ConstructionRay(p1, p2)
        ray2 = ConstructionRay(p2, p3)
        assert ray1.is_horizontal is True
        assert ray2.is_horizontal is True
        assert ray1.is_parallel(ray2) is True
        assert (
            math.isclose(ray1.slope, ray2.slope) is False
        ), "Only slope testing is not sufficient"



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import string
from ezdxf.r12strict import R12NameTranslator


@pytest.fixture
def tr() -> R12NameTranslator:
    return R12NameTranslator()


def test_translate_empty_string(tr):
    assert tr.translate("") == ""
    assert tr.translate("") == ""
    assert len(tr.used_r12_names) == 1
    assert len(tr.translated_names) == 1


def test_translate_uppercase_letters(tr):
    assert tr.translate(string.ascii_uppercase) == string.ascii_uppercase


def test_translate_lowercase_letters(tr):
    assert tr.translate(string.ascii_lowercase) == string.ascii_uppercase


def test_translate_digits(tr):
    assert tr.translate(string.digits) == string.digits


def test_translate_special_chars(tr):
    assert tr.translate("$_-") == "$_-"


def test_asterix_is_allowed_as_first_letter(tr):
    assert tr.translate("*Active") == "*ACTIVE"


def test_translate_punctuation(tr):
    """Punctuations are not supported"""
    assert tr.translate(".,;:!?\\/*") == "_________"


def test_translation_spaces(tr):
    """Spaces are not supported"""
    assert tr.translate("  ") == "__"


def test_translate_invalid_chars_to_underscores(tr):
    assert tr.translate("+abc;") == "_ABC_"


def test_get_same_translations_for_same_names(tr):
    assert tr.translate("+abc;") == "_ABC_"
    assert tr.translate("+abc;") == "_ABC_", "expected the same translation"
    assert len(tr.used_r12_names) == 1
    assert len(tr.translated_names) == 1


def test_unique_names_for_same_r12_translations(tr):
    assert tr.translate("+abc") == "_ABC"
    assert tr.translate("#abc") == "_ABC0"  # basic translation is also _ABC


def test_names_case_insensitive(tr):
    assert tr.translate("Continuous") == "CONTINUOUS"
    assert tr.translate("CONTINUOUS") == "CONTINUOUS"
    assert len(tr.used_r12_names) == 1
    assert len(tr.translated_names) == 1


def test_limit_length_of_long_names(tr):
    name = string.ascii_letters + string.digits
    assert len(tr.translate(name)) == 31


def test_translate_long_names(tr):
    name0 = string.ascii_letters + string.digits
    name1 = name0 + "#"
    assert tr.translate(name0) != tr.translate(name1)
    assert tr.translate(name0) == tr.translate(name0)
    assert tr.translate(name1) == tr.translate(name1)
    assert len(tr.used_r12_names) == 2
    assert len(tr.translated_names) == 2


if __name__ == '__main__':
    pytest.main([__file__])



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import cast, Sequence
from ezdxf.entities import DXFEntity, Body as DXFBody
from ezdxf.lldxf import const
from .entities import Body, export_sat, export_sab, load


def export_dxf(entity: DXFEntity, bodies: Sequence[Body]):
    """Store the :term:`ACIS` bodies in the given DXF entity. This is the
    recommended way to set ACIS data of DXF entities.

    The DXF entity has to be an ACIS based entity and inherit from
    :class:`ezdxf.entities.Body`. The entity has to be bound to a valid DXF
    document and the DXF version of the document has to be DXF R2000 or newer.

    Raises:
        DXFTypeError: invalid DXF entity type
        DXFValueError: invalid DXF document
        DXFVersionError: invalid DXF version

    """
    if not isinstance(entity, DXFBody):
        raise const.DXFTypeError(f"invalid DXF entity {entity.dxftype()}")
    body = cast(DXFBody, entity)
    doc = entity.doc
    if doc is None:
        raise const.DXFValueError("a valid DXF document is required")
    dxfversion = doc.dxfversion
    if dxfversion < const.DXF2000:
        raise const.DXFVersionError(f"invalid DXF version {dxfversion}")

    if dxfversion < const.DXF2013:
        body.sat = export_sat(bodies)
    else:
        body.sab = export_sab(bodies)


def load_dxf(entity: DXFEntity) -> list[Body]:
    """Load the :term:`ACIS` bodies from the given DXF entity. This is the
    recommended way to load ACIS data.

    The DXF entity has to be an ACIS based entity and inherit from
    :class:`ezdxf.entities.Body`. The entity has to be bound to a valid DXF
    document and the DXF version of the document has to be DXF R2000 or newer.

    Raises:
        DXFTypeError: invalid DXF entity type
        DXFValueError: invalid DXF document
        DXFVersionError: invalid DXF version

    .. warning::

        Only a limited count of :term:`ACIS` entities is supported, all
        unsupported entities are loaded as ``NONE_ENTITY`` and their data is
        lost. Exporting such ``NONE_ENTITIES`` will raise an :class:`ExportError`
        exception.

        To emphasize that again: **It is not possible to load and re-export
        arbitrary ACIS data!**

    """

    if not isinstance(entity, DXFBody):
        raise const.DXFTypeError(f"invalid DXF entity {entity.dxftype()}")
    body = cast(DXFBody, entity)
    doc = entity.doc
    if doc is None:
        raise const.DXFValueError("a valid DXF document is required")
    dxfversion = doc.dxfversion
    if dxfversion < const.DXF2000:
        raise const.DXFVersionError(f"invalid DXF version {dxfversion}")
    if body.has_binary_data:
        binary_data = body.sab
        if binary_data:
            return load(binary_data)
    else:
        text = body.sat
        if text:
            return load(text)
    return []



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math.clustering import dbscan
from ezdxf.render import forms


def make_cluster(size, dx, dy, dz):
    return forms.cube().scale_uniform(size).translate(dx, dy, dz).vertices


def test_two_simple_cluster():
    c1 = make_cluster(1, 0, 0, 0)
    c2 = make_cluster(1.5, 5, 0, 0)
    points = list(c1)
    points.extend(c2)
    cluster = dbscan(points, radius=2)
    assert len(cluster) == 2
    cluster.sort()
    assert set(cluster[0]) == set(c1)
    assert set(cluster[1]) == set(c2)


def test_cluster_noise():
    cluster = dbscan(make_cluster(10, 0, 0, 0), radius=2)
    assert len(cluster) == 8  # all noise


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import os
import ezdxf.recover

BASEDIR = os.path.dirname(__file__)
DATADIR = "data"
FILENAME = os.path.join(BASEDIR, DATADIR , "R12_with_trash_beyond_EOF.dxf")


def test_recover_ignores_junk_beyond_EOF():
    doc, auditor = ezdxf.recover.readfile(FILENAME)
    assert len(auditor.errors) == 0
    assert len(auditor.fixes) == 0
    msp = doc.modelspace()
    assert msp[0].dxftype() == "CIRCLE"


def test_readfile_ignores_junk_beyond_EOF():
    doc = ezdxf.readfile(FILENAME)
    msp = doc.modelspace()
    assert msp[0].dxftype() == "CIRCLE"



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import pytest
from ezdxf.entities import DXFEntity
from ezdxf import reorder


@pytest.fixture()
def entities():
    return [
        DXFEntity.new(handle="A"),
        DXFEntity.new(handle="D"),
        DXFEntity.new(handle="B"),
        DXFEntity.new(handle="C"),
    ]


def test_ascending_sort_order(entities):
    handles = [e.dxf.handle for e in reorder.ascending(entities)]
    assert handles == ["A", "B", "C", "D"]


def test_mapped_ascending_sort_order(entities):
    handles = [e.dxf.handle for e in reorder.ascending(entities, {"B": "D"})]
    # B has the same sort handle as D and D & B should show up in source order
    assert handles == ["A", "C", "D", "B"]


def test_mapping_to_0_ascending_sort_order(entities):
    handles = [e.dxf.handle for e in reorder.ascending(entities, {"A": "0"})]
    assert handles == [
        "B",
        "C",
        "D",
        "A",
    ], 'Expected "A" mapped to "0" as last element'


def test_full_mapped_ascending_sort_order(entities):
    handles = [
        e.dxf.handle
        for e in reorder.ascending(
            entities,
            {
                "A": "A",
                "B": "A",
                "C": "A",
                "D": "A",
            },
        )
    ]
    assert handles == ["A", "D", "B", "C"], "Expected the source entity order"


def test_descending_sort_order(entities):
    handles = [e.dxf.handle for e in reorder.descending(entities)]
    assert handles == ["D", "C", "B", "A"]


def test_mapping_to_0_descending_sort_order(entities):
    handles = [e.dxf.handle for e in reorder.descending(entities, {"A": "0"})]
    assert handles == [
        "A",
        "D",
        "C",
        "B",
    ], 'Expected "A" mapped to "0" as first element'


def test_full_mapped_descending_sort_order(entities):
    handles = [
        e.dxf.handle
        for e in reorder.descending(
            entities,
            {
                "A": "A",
                "B": "A",
                "C": "A",
                "D": "A",
            },
        )
    ]
    assert handles == [
        "C",
        "B",
        "D",
        "A",
    ], "Expected the reversed source entity order"


if __name__ == "__main__":
    pytest.main([__file__])



# cython: language_level=3
# Source: https://github.com/mapbox/earcut
# License: ISC License (MIT compatible)
#
# Copyright (c) 2016, Mapbox
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.
#
# Cython implementation of module ezdxf.math._mapbox_earcut.py
# Copyright (c) 2022, Manfred Moitzi
# License: MIT License
# type: ignore
from libc.math cimport fmin, fmax, fabs, INFINITY


cdef class Node:
    cdef:
        int i
        double x
        double y
        int z
        bint steiner
        object point
        Node prev
        Node next
        Node prev_z
        Node next_z

    def __cinit__(self, int i, object point):
        self.i = i
        self.point = point
        self.x = point.x
        self.y = point.y
        self.z = 0
        self.steiner = False
        self.prev = None
        self.next = None
        self.prev_z = None
        self.next_z = None

    cdef bint equals(self, Node other):
        return self.x == other.x and self.y == other.y

def node_key(Node node):
    return node.x, node.y

def earcut(list exterior, list holes):
    """Implements a modified ear slicing algorithm, optimized by z-order
    curve hashing and extended to handle holes, twisted polygons, degeneracies
    and self-intersections in a way that doesn't guarantee correctness of
    triangulation, but attempts to always produce acceptable results for
    practical data.

    Source: https://github.com/mapbox/earcut

    Args:
        exterior: outer path as list of points as objects which provide a
            `x`- and a `y`-attribute
        holes: list of holes, each hole is list of points, a hole with
            a single points is a Steiner point

    Returns:
        Returns a list of triangles, each triangle is a tuple of three points,
        the output points are the same objects as the input points.

    """
    cdef:
        Node outer_node
        list triangles = []
        double max_x, max_y, x, y
        double min_x = 0.0
        double min_y = 0.0
        double inv_size = 0.0

    if not exterior:
        return triangles

    outer_node = linked_list(exterior, 0, ccw=True)
    if outer_node is None or outer_node.next is outer_node.prev:
        return triangles

    if holes:
        outer_node = eliminate_holes(holes, len(exterior), outer_node)

    # if the shape is not too simple, we'll use z-order curve hash later
    # calculate polygon bbox
    if len(exterior) > 80:
        min_x = max_x = exterior[0].x
        min_y = max_y = exterior[0].y
        for point in exterior:
            x = point.x
            y = point.y
            min_x = fmin(min_x, x)
            min_y = fmin(min_y, y)
            max_x = fmax(max_x, x)
            max_y = fmax(max_y, y)

        # min_x, min_y and inv_size are later used to transform coords into
        # integers for z-order calculation
        inv_size = fmax(max_x - min_x, max_y - min_y)
        inv_size = 32767 / inv_size if inv_size != 0 else 0

    earcut_linked(outer_node, triangles, min_x, min_y, inv_size, 0)
    return triangles


cdef Node linked_list(list points, int start, bint ccw):
    """Create a circular doubly linked list from polygon points in the specified
    winding order
    """
    cdef:
        Node last = None
        int end

    if ccw is (signed_area(points) < 0):
        for point in points:
            last = insert_node(start, point, last)
            start += 1
    else:
        end = start + len(points)
        for point in reversed(points):
            last = insert_node(end, point, last)
            end -= 1

    # open polygon: where the 1st vertex is not coincident with the last vertex
    if last and last.equals(last.next):
        remove_node(last)
        last = last.next
    return last


cdef double signed_area(list points):
    cdef:
        double s = 0.0
        double point_x, prev_x, point_y, prev_y
    if not len(points):
        return s
    prev = points[-1]
    prev_x = prev.x
    prev_y = prev.y
    for point in points:
        point_x = point.x
        point_y = point.y
        s += (point_x - prev_x) * (point_y + prev_y)
        prev_x = point_x
        prev_y = point_y
    # s < 0 is counter-clockwise
    # s > 0 is clockwise
    return s


cdef double area(Node p, Node q, Node r):
    """Returns signed area of a triangle"""
    return (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y)


cdef bint is_valid_diagonal(Node a, Node b):
    """Check if a diagonal between two polygon nodes is valid (lies in polygon
    interior)
    """
    return (
        a.next.i != b.i
        and a.prev.i != b.i
        and not intersects_polygon(a, b)  # doesn't intersect other edges
        and (
            locally_inside(a, b)
            and locally_inside(b, a)
            and middle_inside(a, b)
            and (
                area(a.prev, a, b.prev) or area(a, b.prev, b)
            )  # does not create opposite-facing sectors
            or a.equals(b)
            and area(a.prev, a, a.next) > 0
            and area(b.prev, b, b.next) > 0
        )  # special zero-length case
    )


cdef bint intersects_polygon(Node a, Node b):
    """Check if a polygon diagonal intersects any polygon segments"""
    cdef Node p = a
    while True:
        if (
            p.i != a.i
            and p.next.i != a.i
            and p.i != b.i
            and p.next.i != b.i
            and intersects(p, p.next, a, b)
        ):
            return True
        p = p.next
        if p is a:
            break
    return False


cdef int sign(double num):
    if num < 0.0:
        return -1
    if num > 0.0:
        return 1
    return 0


cdef bint on_segment(p: Node, q: Node, r: Node):
    return fmax(p.x, r.x) >= q.x >= fmin(p.x, r.x) and fmax(
        p.y, r.y
    ) >= q.y >= fmin(p.y, r.y)


cdef bint intersects(Node p1, Node q1, Node p2, Node q2):
    """check if two segments intersect"""
    cdef:
        int o1 = sign(area(p1, q1, p2))
        int o2 = sign(area(p1, q1, q2))
        int o3 = sign(area(p2, q2, p1))
        int o4 = sign(area(p2, q2, q1))

    if o1 != o2 and o3 != o4:
        return True  # general case

    if o1 == 0 and on_segment(p1, p2, q1):
        return True  # p1, q1 and p2 are collinear and p2 lies on p1q1
    if o2 == 0 and on_segment(p1, q2, q1):
        return True  # p1, q1 and q2 are collinear and q2 lies on p1q1
    if o3 == 0 and on_segment(p2, p1, q2):
        return True  # p2, q2 and p1 are collinear and p1 lies on p2q2
    if o4 == 0 and on_segment(p2, q1, q2):
        return True  # p2, q2 and q1 are collinear and q1 lies on p2q2
    return False


cdef Node insert_node(int i, point, Node last):
    """create a node and optionally link it with previous one (in a circular
    doubly linked list)
    """
    cdef Node p = Node(i, point)

    if last is None:
        p.prev = p
        p.next = p
    else:
        p.next = last.next
        p.prev = last
        last.next.prev = p
        last.next = p
    return p


cdef remove_node(Node p):
    p.next.prev = p.prev
    p.prev.next = p.next

    if p.prev_z is not None:
        p.prev_z.next_z = p.next_z
    if p.next_z is not None:
        p.next_z.prev_z = p.prev_z


cdef Node eliminate_holes(list holes, int start, Node outer_node):
    """link every hole into the outer loop, producing a single-ring polygon
    without holes
    """
    cdef:
        list queue = []
        list hole
    for hole in holes:
        if len(hole) < 1:  # skip empty holes
            continue
        # hole vertices in clockwise order
        _list = linked_list(hole, start, ccw=False)
        if _list is _list.next:
            _list.steiner = True
        start += len(hole)
        queue.append(get_leftmost(_list))
    queue.sort(key=node_key)

    # process holes from left to right
    for hole_ in queue:
        outer_node = eliminate_hole(hole_, outer_node)
    return outer_node


cdef Node eliminate_hole(Node hole, Node outer_node):
    """Find a bridge between vertices that connects hole with an outer ring and
    link it
    """
    cdef:
        Node bridge = find_hole_bridge(hole, outer_node)
        Node bridge_reverse

    if bridge is None:
        return outer_node

    bridge_reverse = split_polygon(bridge, hole)

    # filter collinear points around the cuts
    filter_points(bridge_reverse, bridge_reverse.next)
    return filter_points(bridge, bridge.next)


cdef Node filter_points(Node start, Node end = None):
    """eliminate colinear or duplicate points"""
    cdef:
        Node p
        bint again

    if start is None:
        return start
    if end is None:
        end = start

    p = start
    while True:
        again = False
        if not p.steiner and (
            p.equals(p.next) or area(p.prev, p, p.next) == 0
        ):
            remove_node(p)
            p = end = p.prev
            if p is p.next:
                break
            again = True
        else:
            p = p.next
        if not (again or p is not end):
            break
    return end


# main ear slicing loop which triangulates a polygon (given as a linked list)
cdef earcut_linked(
    Node ear,
    list triangles,
    double min_x,
    double min_y,
    double inv_size,
    int pass_,
):
    cdef:
        Node stop, prev, next
        bint _is_ear

    if ear is None:
        return

    # interlink polygon nodes in z-order
    if not pass_ and inv_size:
        index_curve(ear, min_x, min_y, inv_size)

    stop = ear

    # iterate through ears, slicing them one by one
    while ear.prev is not ear.next:
        prev = ear.prev
        next = ear.next

        _is_ear = (
            is_ear_hashed(ear, min_x, min_y, inv_size)
            if inv_size
            else is_ear(ear)
        )
        if _is_ear:
            # cut off the triangle
            triangles.append((prev.point, ear.point, next.point))
            remove_node(ear)

            # skipping the next vertex leads to less sliver triangles
            ear = next.next
            stop = next.next
            continue

        ear = next

        # if we looped through the whole remaining polygon and can't find any more ears
        if ear is stop:
            # try filtering points and slicing again
            if not pass_:
                earcut_linked(
                    filter_points(ear),
                    triangles,
                    min_x,
                    min_y,
                    inv_size,
                    1,
                )

            # if this didn't work, try curing all small self-intersections locally
            elif pass_ == 1:
                ear = cure_local_intersections(filter_points(ear), triangles)
                earcut_linked(ear, triangles, min_x, min_y, inv_size, 2)

            # as a last resort, try splitting the remaining polygon into two
            elif pass_ == 2:
                split_ear_cut(ear, triangles, min_x, min_y, inv_size)
            break


cdef bint is_ear(Node ear):
    """check whether a polygon node forms a valid ear with adjacent nodes"""
    cdef:
        Node a = ear.prev
        Node b = ear
        Node c = ear.next
        Node p
        double x0, x1, y0, y1

    if area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    # now make sure we don't have other points inside the potential ear
    # triangle bbox
    x0 = fmin(a.x, fmin(b.x, c.x))
    x1 = fmax(a.x, fmax(b.x, c.x))
    y0 = fmin(a.y, fmin(b.y, c.y))
    y1 = fmax(a.y, fmax(b.y, c.y))
    p = c.next

    while p is not a:
        if (
            x0 <= p.x <= x1
            and y0 <= p.y <= y1
            and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y)
            and area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.next

    return True


cdef bint is_ear_hashed(Node ear, double min_x, double min_y, double inv_size):
    cdef:
        Node a = ear.prev
        Node b = ear
        Node c = ear.next
        double x0, x1, y0, y1, min_z, max_z
        Node p, n

    if area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    # triangle bbox
    x0 = fmin(a.x, fmin(b.x, c.x))
    x1 = fmax(a.x, fmax(b.x, c.x))
    y0 = fmin(a.y, fmin(b.y, c.y))
    y1 = fmax(a.y, fmax(b.y, c.y))

    # z-order range for the current triangle bbox;
    min_z = z_order(x0, y0, min_x, min_y, inv_size)
    max_z = z_order(x1, y1, min_x, min_y, inv_size)

    p = ear.prev_z
    n = ear.next_z

    # look for points inside the triangle in both directions
    while p and p.z >= min_z and n and n.z <= max_z:
        if (
            x0 <= p.x <= x1
            and y0 <= p.y <= y1
            and p is not a
            and p is not c
            and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y)
            and area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.prev_z

        if (
            x0 <= n.x <= x1
            and y0 <= n.y <= y1
            and n is not a
            and n is not c
            and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, n.x, n.y)
            and area(n.prev, n, n.next) >= 0
        ):
            return False
        n = n.next_z

    # look for remaining points in decreasing z-order
    while p and p.z >= min_z:
        if (
            x0 <= p.x <= x1
            and y0 <= p.y <= y1
            and p is not a
            and p is not c
            and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y)
            and area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.prev_z

    # look for remaining points in increasing z-order
    while n and n.z <= max_z:
        if (
            x0 <= n.x <= x1
            and y0 <= n.y <= y1
            and n is not a
            and n is not c
            and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, n.x, n.y)
            and area(n.prev, n, n.next) >= 0
        ):
            return False
        n = n.next_z
    return True


cdef Node get_leftmost(Node start):
    """Find the leftmost node of a polygon ring"""
    cdef:
        Node p = start
        Node leftmost = start

    while True:
        if p.x < leftmost.x or (p.x == leftmost.x and p.y < leftmost.y):
            leftmost = p
        p = p.next
        if p is start:
            break
    return leftmost


cdef bint point_in_triangle(
    double ax,
    double ay,
    double bx,
    double by_,
    double cx,
    double cy,
    double px,
    double py,
):
    """Check if a point lies within a convex triangle"""
    return (
        (cx - px) * (ay - py) >= (ax - px) * (cy - py)
        and (ax - px) * (by_ - py) >= (bx - px) * (ay - py)
        and (bx - px) * (cy - py) >= (cx - px) * (by_ - py)
    )


cdef bint sector_contains_sector(Node m, Node p):
    """Whether sector in vertex m contains sector in vertex p in the same
    coordinates.
    """
    return area(m.prev, m, p.prev) < 0 and area(p.next, m, m.next) < 0


cdef index_curve(Node start, double min_x, double min_y, double inv_size):
    """Interlink polygon nodes in z-order"""
    cdef Node p = start
    while True:
        if p.z == 0:
            p.z = z_order(p.x, p.y, min_x, min_y, inv_size)
        p.prev_z = p.prev
        p.next_z = p.next
        p = p.next
        if p is start:
            break

    p.prev_z.next_z = None
    p.prev_z = None

    sort_linked(p)


cdef int z_order(
    double x0, double y0, double min_x, double min_y, double inv_size
):
    """Z-order of a point given coords and inverse of the longer side of data
    bbox.
    """
    # coords are transformed into non-negative 15-bit integer range
    cdef:
        int x = int((x0 - min_x) * inv_size)
        int y = int ((y0 - min_y) * inv_size)

    x = (x | (x << 8)) & 0x00FF00FF
    x = (x | (x << 4)) & 0x0F0F0F0F
    x = (x | (x << 2)) & 0x33333333
    x = (x | (x << 1)) & 0x55555555

    y = (y | (y << 8)) & 0x00FF00FF
    y = (y | (y << 4)) & 0x0F0F0F0F
    y = (y | (y << 2)) & 0x33333333
    y = (y | (y << 1)) & 0x55555555

    return x | (y << 1)


# Simon Tatham's linked list merge sort algorithm
# http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.html
cdef Node sort_linked(Node head):
    cdef:
        int in_size = 1
        int num_merges, p_size, q_size, i
        Node tail, p, q, e

    while True:
        p = head
        head = None
        tail = None
        num_merges = 0
        while p:
            num_merges += 1
            q = p
            p_size = 0
            for i in range(in_size):
                p_size += 1
                q = q.next_z
                if not q:
                    break
            q_size = in_size
            while p_size > 0 or (q_size > 0 and q):
                if p_size != 0 and (q_size == 0 or not q or p.z <= q.z):
                    e = p
                    p = p.next_z
                    p_size -= 1
                else:
                    e = q
                    q = q.next_z
                    q_size -= 1

                if tail:
                    tail.next_z = e
                else:
                    head = e
                e.prev_z = tail
                tail = e
            p = q
        tail.next_z = None
        in_size *= 2
        if num_merges <= 1:
            break
    return head


cdef Node split_polygon(Node a, Node b):
    """Link two polygon vertices with a bridge.

    If the vertices belong to the same ring, it splits polygon into two.
    If one belongs to the outer ring and another to a hole, it merges it into a
    single ring.
    """
    cdef :
        Node a2 = Node(a.i, a.point)
        Node b2 = Node(b.i, b.point)
        Node an = a.next
        Node bp = b.prev

    a.next = b
    b.prev = a

    a2.next = an
    an.prev = a2

    b2.next = a2
    a2.prev = b2

    bp.next = b2
    b2.prev = bp

    return b2


# go through all polygon nodes and cure small local self-intersections
cdef Node cure_local_intersections(Node start, list triangles):
    cdef:
        Node p = start
        Node a, b
    while True:
        a = p.prev
        b = p.next.next

        if (
            not a.equals(b)
            and intersects(a, p, p.next, b)
            and locally_inside(a, b)
            and locally_inside(b, a)
        ):
            triangles.append((a.point, p.point, b.point))
            # remove two nodes involved
            remove_node(p)
            remove_node(p.next)
            p = start = b

        p = p.next
        if p is start:
            break
    return filter_points(p)


cdef split_ear_cut(
    Node start,
    list triangles,
    double min_x,
    double min_y,
    double inv_size,
):
    """Try splitting polygon into two and triangulate them independently"""
    # look for a valid diagonal that divides the polygon into two
    cdef:
        Node a = start
        Node b, c

    while True:
        b = a.next.next
        while b is not a.prev:
            if a.i != b.i and is_valid_diagonal(a, b):
                # split the polygon in two by the diagonal
                c = split_polygon(a, b)

                # filter colinear points around the cuts
                a = filter_points(a, a.next)
                c = filter_points(c, c.next)

                # run earcut on each half
                earcut_linked(a, triangles, min_x, min_y, inv_size, 0)
                earcut_linked(c, triangles, min_x, min_y, inv_size, 0)
                return
            b = b.next
        a = a.next
        if a is start:
            break


# David Eberly's algorithm for finding a bridge between hole and outer polygon
cdef Node find_hole_bridge(Node hole, Node outer_node):
    cdef:
        Node p = outer_node
        Node m = None
        Node stop
        double hx = hole.x
        double hy = hole.y
        double qx = -INFINITY
        double mx, my, tan_min, tan

    # find a segment intersected by a ray from the hole's leftmost point to the left;
    # segment's endpoint with lesser x will be potential connection point
    while True:
        if p.y >= hy >= p.next.y != p.y:
            x = p.x + (hy - p.y) * (p.next.x - p.x) / (p.next.y - p.y)
            if hx >= x > qx:
                qx = x
                m = p if p.x < p.next.x else p.next
                if x == hx:
                    # hole touches outer segment; pick leftmost endpoint
                    return m
        p = p.next
        if p is outer_node:
            break

    if m is None:
        return None

    # look for points inside the triangle of hole point, segment intersection and endpoint;
    # if there are no points found, we have a valid connection;
    # otherwise choose the point of the minimum angle with the ray as connection point
    stop = m
    mx = m.x
    my = m.y
    tan_min = INFINITY
    p = m

    while True:
        if (
            hx >= p.x >= mx
            and hx != p.x
            and point_in_triangle(
                hx if hy < my else qx,
                hy,
                mx,
                my,
                qx if hy < my else hx,
                hy,
                p.x,
                p.y,
            )
        ):
            tan = fabs(hy - p.y) / (hx - p.x)  # tangential
            if locally_inside(p, hole) and (
                tan < tan_min
                or (
                    tan == tan_min
                    and (
                        p.x > m.x
                        or (p.x == m.x and sector_contains_sector(m, p))
                    )
                )
            ):
                m = p
                tan_min = tan
        p = p.next
        if p is stop:
            break
    return m


cdef bint locally_inside(Node a, Node b):
    """Check if a polygon diagonal is locally inside the polygon"""
    return (
        area(a, b, a.next) >= 0 and area(a, a.prev, b) >= 0
        if area(a.prev, a, a.next) < 0
        else area(a, b, a.prev) < 0 or area(a, a.next, b) < 0
    )


cdef bint middle_inside(Node a, Node b):
    """Check if the middle point of a polygon diagonal is inside the polygon"""
    cdef:
        Node p = a
        bint inside = False
        double px = (a.x + b.x) / 2
        double py = (a.y + b.y) / 2

    while True:
        if (
            ((p.y > py) != (p.next.y > py))
            and p.next.y != p.y
            and (px < (p.next.x - p.x) * (py - p.y) / (p.next.y - p.y) + p.x)
        ):
            inside = not inside
        p = p.next
        if p is a:
            break
    return inside
